"""
These commands are used to build Salt packages.
"""

# pylint: disable=resource-leakage,broad-except,3rd-party-module-not-gated
from __future__ import annotations

import fnmatch
import gzip
import hashlib
import json
import logging
import os
import pathlib
import shutil
import sys
import tarfile
import tempfile

import yaml
from ptscripts import Context, command_group
from ptscripts.models import VirtualEnvPipConfig

import tools.utils

log = logging.getLogger(__name__)

# Define the command group
pkg = command_group(name="pkg", help="Packaging Related Commands", description=__doc__)


class Recompress:
    """
    Helper class to re-compress a ``.tag.gz`` file to make it reproducible.
    """

    def __init__(self, mtime):
        self.mtime = int(mtime)

    def tar_reset(self, tarinfo):
        """
        Reset user, group, mtime, and mode to create reproducible tar.
        """
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = "root"
        tarinfo.mtime = self.mtime
        if tarinfo.type == tarfile.DIRTYPE:
            tarinfo.mode = 0o755
        else:
            tarinfo.mode = 0o644
        if tarinfo.pax_headers:
            raise ValueError(tarinfo.name, tarinfo.pax_headers)
        return tarinfo

    def recompress(self, targz):
        """
        Re-compress the passed path.
        """
        tempd = pathlib.Path(tempfile.mkdtemp()).resolve()
        d_src = tempd.joinpath("src")
        d_src.mkdir()
        d_tar = tempd.joinpath(targz.stem)
        d_targz = tempd.joinpath(targz.name)
        with tarfile.open(d_tar, "w|") as wfile:
            with tarfile.open(targz, "r:gz") as rfile:
                rfile.extractall(d_src)  # nosec
                extracted_dir = next(pathlib.Path(d_src).iterdir())
                for name in sorted(extracted_dir.rglob("*")):
                    wfile.add(
                        str(name),
                        filter=self.tar_reset,
                        recursive=False,
                        arcname=str(name.relative_to(d_src)),
                    )

        with open(d_tar, "rb") as rfh:
            with gzip.GzipFile(
                fileobj=open(d_targz, "wb"), mode="wb", filename="", mtime=self.mtime
            ) as gz:  # pylint: disable=invalid-name
                while True:
                    chunk = rfh.read(1024)
                    if not chunk:
                        break
                    gz.write(chunk)
        targz.unlink()
        shutil.move(str(d_targz), str(targz))


@pkg.command(
    name="set-salt-version",
    arguments={
        "salt_version": {
            "help": (
                "The salt version to write to 'salt/_version.txt'. If not passed "
                "it will be discovered by running 'python3 salt/version.py'."
            ),
            "nargs": "?",
            "default": None,
        },
        "overwrite": {
            "help": "Overwrite 'salt/_version.txt' if it already exists",
        },
        "validate_version": {
            "help": "Validate, and normalize, the passed Salt Version",
        },
        "release": {
            "help": "When true, also update salt/versions.py to set the version as released",
        },
    },
)
def set_salt_version(
    ctx: Context,
    salt_version: str,
    overwrite: bool = False,
    validate_version: bool = False,
    release: bool = False,
):
    """
    Write the Salt version to 'salt/_version.txt'
    """
    salt_version_file = tools.utils.REPO_ROOT / "salt" / "_version.txt"
    if salt_version_file.exists():
        if not overwrite:
            ctx.error("The 'salt/_version.txt' file already exists")
            ctx.exit(1)
        salt_version_file.unlink()
    if salt_version is None:
        if not tools.utils.REPO_ROOT.joinpath(".git").exists():
            ctx.error(
                "Apparently not running from a Salt repository checkout. "
                "Unable to discover the Salt version."
            )
            ctx.exit(1)
            ctx.info("Discovering the Salt version...")
        ret = ctx.run(shutil.which("python3"), "salt/version.py", capture=True)
        salt_version = ret.stdout.strip().decode()
        ctx.info(f"Discovered Salt version: {salt_version!r}")
    elif validate_version:
        ctx.info(f"Validating and normalizing the salt version {salt_version!r}...")
        with ctx.virtualenv(
            name="set-salt-version",
            config=VirtualEnvPipConfig(
                pip_requirement="pip>=24.2",
                requirements_files=[
                    tools.utils.REPO_ROOT / "requirements" / "base.txt",
                ],
            ),
        ) as venv:
            code = f"""
            import sys
            import salt.version
            parsed_version = salt.version.SaltStackVersion.parse("{salt_version}")
            if parsed_version.name is None:
                # When we run out of names, or we stop supporting version names
                # we'll need to remove this version check.
                print("'{{}}' is not a valid Salt Version.".format(parsed_version), file=sys.stderr, flush=True)
                sys.exit(1)
            sys.stdout.write(str(parsed_version))
            sys.stdout.flush()
            """
            ret = venv.run_code(code, capture=True, check=False)
            if ret.returncode:
                ctx.error(ret.stderr.decode())
                ctx.exit(ret.returncode)
            salt_version = ret.stdout.strip().decode()

    if not tools.utils.REPO_ROOT.joinpath("salt").is_dir():
        ctx.error(
            "The path 'salt/' is not a directory. Unable to write 'salt/_version.txt'"
        )
        ctx.exit(1)

    try:
        tools.utils.REPO_ROOT.joinpath("salt/_version.txt").write_text(
            salt_version, encoding="utf-8"
        )
    except Exception as exc:
        ctx.error(f"Unable to write 'salt/_version.txt': {exc}")
        ctx.exit(1)

    ctx.info(f"Successfuly wrote {salt_version!r} to 'salt/_version.txt'")

    version_instance = tools.utils.Version(salt_version)
    if release and not version_instance.is_prerelease:
        with open(
            tools.utils.REPO_ROOT / "salt" / "version.py", "r+", encoding="utf-8"
        ) as rwfh:
            contents = rwfh.read()
            match = f"info=({version_instance.major}, {version_instance.minor}))"
            if match in contents:
                contents = contents.replace(
                    match,
                    f"info=({version_instance.major}, {version_instance.minor}),  released=True)",
                )
                rwfh.seek(0)
                rwfh.write(contents)
                rwfh.truncate()

                ctx.info(
                    f"Successfuly marked {salt_version!r} as released in 'salt/version.py'"
                )

    gh_env_file = os.environ.get("GITHUB_ENV", None)
    if gh_env_file is not None:
        variable_text = f"SALT_VERSION={salt_version}"
        ctx.info(f"Writing '{variable_text}' to '$GITHUB_ENV' file:", gh_env_file)
        with open(gh_env_file, "w", encoding="utf-8") as wfh:
            wfh.write(f"{variable_text}\n")

    gh_output_file = os.environ.get("GITHUB_OUTPUT", None)
    if gh_output_file is not None:
        variable_text = f"salt-version={salt_version}"
        ctx.info(f"Writing '{variable_text}' to '$GITHUB_OUTPUT' file:", gh_output_file)
        with open(gh_output_file, "w", encoding="utf-8") as wfh:
            wfh.write(f"{variable_text}\n")

    ctx.exit(0)


@pkg.command(
    name="pre-archive-cleanup",
    arguments={
        "cleanup_path": {
            "help": (
                "The salt version to write to 'salt/_version.txt'. If not passed "
                "it will be discovered by running 'python3 salt/version.py'."
            ),
            "metavar": "PATH_TO_CLEANUP",
        },
        "pkg": {
            "help": "Perform extended, pre-packaging cleanup routines",
        },
    },
)
def pre_archive_cleanup(ctx: Context, cleanup_path: str, pkg: bool = False):
    """
    Clean the provided path of paths that should not be included in the archive.

    For example:

        * `__pycache__` directories
        * `*.pyc` files
        * `*.pyo` files

    When running on Windows and macOS, some additional cleanup is also done.
    """
    with open(
        str(tools.utils.REPO_ROOT / "pkg" / "common" / "env-cleanup-rules.yml"),
        encoding="utf-8",
    ) as rfh:
        patterns = yaml.safe_load(rfh.read())

    if pkg:
        patterns = patterns["pkg"]
    else:
        patterns = patterns["ci"]

    if sys.platform.lower().startswith("win"):
        patterns = patterns["windows"]
    elif sys.platform.lower().startswith("darwin"):
        patterns = patterns["darwin"]
    else:
        patterns = patterns["linux"]

    def unnest_lists(patterns):
        if isinstance(patterns, list):
            for pattern in patterns:
                yield from unnest_lists(pattern)
        else:
            yield patterns

    exclude_patterns = set()
    for pattern in unnest_lists(patterns["exclude_patterns"]):
        exclude_patterns.add(pattern)

    dir_patterns = set()
    for pattern in unnest_lists(patterns["dir_patterns"]):
        dir_patterns.add(pattern)

    file_patterns = set()
    for pattern in unnest_lists(patterns["file_patterns"]):
        file_patterns.add(pattern)

    for root, dirs, files in os.walk(cleanup_path, topdown=True, followlinks=False):
        for dirname in dirs:
            path = pathlib.Path(root, dirname).resolve()
            if not path.exists():
                continue
            match_path = path.as_posix()
            skip_match = False
            for pattern in exclude_patterns:
                if fnmatch.fnmatch(str(match_path), pattern):
                    ctx.info(
                        f"Excluded file: {match_path}; Matching pattern: {pattern!r}"
                    )
                    skip_match = True
                    break
            if skip_match:
                continue
            for pattern in dir_patterns:
                if fnmatch.fnmatch(str(match_path), pattern):
                    ctx.info(
                        f"Deleting directory: {match_path}; Matching pattern: {pattern!r}"
                    )
                    shutil.rmtree(str(path))
                    break
        for filename in files:
            path = pathlib.Path(root, filename).resolve()
            if not path.exists():
                continue
            match_path = path.as_posix()
            skip_match = False
            for pattern in exclude_patterns:
                if fnmatch.fnmatch(str(match_path), pattern):
                    ctx.info(
                        f"Excluded file: {match_path}; Matching pattern: {pattern!r}"
                    )
                    skip_match = True
                    break
            if skip_match:
                continue
            for pattern in file_patterns:
                if fnmatch.fnmatch(str(match_path), pattern):
                    ctx.info(
                        f"Deleting file: {match_path}; Matching pattern: {pattern!r}"
                    )
                    try:
                        os.remove(str(path))
                    except FileNotFoundError:
                        pass
                    break


@pkg.command(
    name="generate-hashes",
    arguments={
        "files": {
            "help": "The files to generate the hashes for.",
            "nargs": "*",
        },
    },
)
def generate_hashes(ctx: Context, files: list[pathlib.Path]):
    """
    Generate "blake2b", "sha512" and "sha3_512" hashes for the passed files.
    """
    for fpath in files:
        ctx.info(f"* Processing {fpath} ...")
        hashes = {}
        for hash_name in ("blake2b", "sha512", "sha3_512"):
            ctx.info(f"   * Calculating {hash_name} ...")
            with fpath.open("rb") as rfh:
                try:
                    digest = hashlib.file_digest(rfh, hash_name)  # type: ignore[attr-defined]
                except AttributeError:
                    # Python < 3.11
                    buf = bytearray(2**18)  # Reusable buffer to reduce allocations.
                    view = memoryview(buf)
                    digest = getattr(hashlib, hash_name)()
                    while True:
                        size = rfh.readinto(buf)
                        if size == 0:
                            break  # EOF
                        digest.update(view[:size])
            digest_file_path = fpath.parent / f"{fpath.name}.{hash_name.upper()}"
            hexdigest = digest.hexdigest()
            ctx.info(f"   * Writing {digest_file_path} ...")
            digest_file_path.write_text(digest.hexdigest())
            hashes[hash_name] = hexdigest
        hashes_json_path = fpath.parent / f"{fpath.name}.json"
        ctx.info(f"   * Writing {hashes_json_path} ...")
        hashes_json_path.write_text(json.dumps(hashes))
    ctx.info("Done")


@pkg.command(
    name="source-tarball",
    venv_config=VirtualEnvPipConfig(
        pip_requirement="pip>=24.2",
        requirements_files=[
            tools.utils.REPO_ROOT / "requirements" / "build.txt",
        ],
    ),
)
def source_tarball(ctx: Context):
    shutil.rmtree("dist/", ignore_errors=True)
    timestamp = ctx.run(
        "git",
        "show",
        "-s",
        "--format=%at",
        "HEAD",
        capture=True,
    ).stdout.strip()
    env = {
        **os.environ,
        **{
            "SOURCE_DATE_EPOCH": str(timestamp),
        },
    }
    ctx.run(
        "python3",
        "-m",
        "build",
        "--sdist",
        str(tools.utils.REPO_ROOT),
        env=env,
        check=True,
    )
    # Recreate sdist to be reproducible
    recompress = Recompress(timestamp)
    for targz in tools.utils.REPO_ROOT.joinpath("dist").glob("*.tar.gz"):
        ctx.info(f"Re-compressing {targz.relative_to(tools.utils.REPO_ROOT)} ...")
        recompress.recompress(targz)
    sha256sum = shutil.which("sha256sum")
    if sha256sum:
        packages = [
            str(pkg.relative_to(tools.utils.REPO_ROOT))
            for pkg in tools.utils.REPO_ROOT.joinpath("dist").iterdir()
        ]
        ctx.run("sha256sum", *packages)
    ctx.run("python3", "-m", "twine", "check", "dist/*", check=True)


@pkg.command(
    name="pypi-upload",
    venv_config=VirtualEnvPipConfig(
        pip_requirement="pip>=24.2",
        requirements_files=[
            tools.utils.REPO_ROOT / "requirements" / "build.txt",
        ],
    ),
    arguments={
        "files": {
            "help": "Files to upload to PyPi",
            "nargs": "*",
        },
        "test": {
            "help": "When true, upload to test.pypi.org instead",
        },
    },
)
def pypi_upload(ctx: Context, files: list[pathlib.Path], test: bool = False):
    ctx.run(
        "python3", "-m", "twine", "check", *[str(fpath) for fpath in files], check=True
    )
    if test is True:
        repository_url = "https://test.pypi.org/legacy/"
    else:
        repository_url = "https://upload.pypi.org/legacy/"
    if "TWINE_USERNAME" not in os.environ:
        os.environ["TWINE_USERNAME"] = "__token__"
    if "TWINE_PASSWORD" not in os.environ:
        ctx.error("The 'TWINE_PASSWORD' variable is not set. Cannot upload.")
        ctx.exit(1)
    cmdline = [
        "twine",
        "upload",
        f"--repository-url={repository_url}",
        "--username=__token__",
    ]
    if test is True:
        cmdline.append("--skip-existing")
    cmdline.extend([str(fpath) for fpath in files])
    ctx.info(f"Running '{' '.join(cmdline)}' ...")
    ret = ctx.run(*cmdline, check=False)
    if ret.returncode:
        ctx.error(ret.stderr.strip().decode())
    ctx.exit(ret.returncode)


@pkg.command(
    name="configure-git",
    arguments={
        "user": {
            "help": "The git global username",
            "required": False,
        },
        "email": {
            "help": "The git global email",
            "required": False,
        },
    },
)
def configure_git(
    ctx: Context,
    user: str = "Salt Project Packaging",
    email: str = "saltproject-packaging@vmware.com",
):
    cwd = pathlib.Path.cwd()
    ctx.info("Setting name and email in git global config")
    ctx.run("git", "config", "--global", "user.name", f"'{user}'")
    ctx.run("git", "config", "--global", "user.email", f"{email}")
    ctx.info(f"Adding {str(cwd)} as a safe directory")
    ctx.run("git", "config", "--global", "--add", "safe.directory", str(cwd))


@pkg.command(
    name="apply-release-patch",
    arguments={
        "patch": {"help": "The git global username"},
        "delete": {
            "help": "Whether to delete the patch after applying",
            "required": False,
        },
    },
)
def apply_release_patch(ctx: Context, patch: pathlib.Path, delete: bool = False):
    patch = patch.resolve()
    ctx.info("Applying the release patch")
    ctx.run("git", "am", "--committer-date-is-author-date", patch.name)
    if delete:
        ctx.info("Deleting the release patch because --delete was passed")
        patch.unlink()



"""
Unit tests for :class:`salt.utils.batch_manager.BatchManager`.

The manager is written so its handlers (``_handle_event``,
``_progress_one``, ``_tick``, ``_handle_new``, ``_handle_stop``,
``_handle_recover``) can be driven directly on an instance whose
``event``, ``local``, and ``output`` dependencies have been replaced
with mocks.  That lets us assert behavior step-by-step without
forking a process, talking to a real event bus, or publishing jobs.
"""

import pytest

import salt.utils.batch_output
import salt.utils.batch_state
from salt.utils.batch_manager import BatchManager, _event_to_return
from tests.support.mock import MagicMock


@pytest.fixture
def opts(tmp_path):
    return {
        "cachedir": str(tmp_path),
        "hash_type": "sha256",
        "sock_dir": str(tmp_path),
        "conf_file": str(tmp_path / "master"),
        "batch_manager_loop_interval": 5,
    }


@pytest.fixture
def manager(opts):
    """
    BatchManager wired up with mocks instead of real event bus / client.

    We skip the ``SignalHandlingProcess.__init__`` side-effects by
    instantiating normally but manually injecting the post-fork
    collaborators so no fork or sock is required.
    """
    mgr = BatchManager(opts)
    mgr.event = MagicMock()
    mgr.local = MagicMock()
    mgr.output = MagicMock()
    mgr.active_batches = set()
    return mgr


def _write_state(opts, minions, jid, batch_size=2, **overrides):
    state = salt.utils.batch_state.create_batch_state(
        {"batch": batch_size, "fun": "test.ping", "tgt": "*"},
        minions,
        jid,
        driver="master",
        now=1000.0,
    )
    state.update(overrides)
    salt.utils.batch_state.write_batch_state(jid, state, opts)
    return state


# ---------------------------------------------------------------------------
# _event_to_return — translation from event bus payload to state-machine input
# ---------------------------------------------------------------------------


class TestEventToReturn:
    def test_maps_return_to_ret(self):
        assert _event_to_return({"return": True, "retcode": 0}) == {
            "ret": True,
            "retcode": 0,
        }

    def test_retcode_default_zero(self):
        assert _event_to_return({"return": "hi"}) == {"ret": "hi", "retcode": 0}

    def test_non_dict_wrapped(self):
        assert _event_to_return("scalar") == {"ret": "scalar", "retcode": 0}


# ---------------------------------------------------------------------------
# _handle_event — dispatch table
# ---------------------------------------------------------------------------


class TestHandleEvent:
    def test_batch_new_calls_handle_new(self, manager):
        manager._handle_new = MagicMock()
        manager._handle_event({"tag": "salt/batch/JID1/new", "data": {}})
        manager._handle_new.assert_called_once_with("JID1")

    def test_batch_stop_calls_handle_stop(self, manager):
        manager._handle_stop = MagicMock()
        manager._handle_event(
            {"tag": "salt/batch/JID1/stop", "data": {"reason": "user"}}
        )
        manager._handle_stop.assert_called_once_with("JID1", {"reason": "user"})

    def test_batch_recover_calls_handle_recover(self, manager):
        manager._handle_recover = MagicMock()
        manager._handle_event({"tag": "salt/batch/JID1/recover", "data": {}})
        manager._handle_recover.assert_called_once_with("JID1")

    def test_job_return_for_adopted_batch_dispatches(self, manager):
        manager.active_batches.add("JID1")
        manager._handle_batch_return = MagicMock()
        manager._handle_event(
            {
                "tag": "salt/job/JID1/ret/web1",
                "data": {"return": True, "retcode": 0},
            }
        )
        manager._handle_batch_return.assert_called_once_with(
            "JID1", "web1", {"return": True, "retcode": 0}
        )

    def test_job_return_for_unknown_jid_is_ignored(self, manager):
        manager._handle_batch_return = MagicMock()
        # Not in active_batches — this is the fast path we don't want to
        # waste I/O on.
        manager._handle_event({"tag": "salt/job/OTHER/ret/web1", "data": {}})
        manager._handle_batch_return.assert_not_called()

    def test_unrelated_event_is_ignored(self, manager):
        manager._handle_batch_return = MagicMock()
        manager._handle_new = MagicMock()
        manager._handle_event({"tag": "salt/auth/accept", "data": {}})
        manager._handle_batch_return.assert_not_called()
        manager._handle_new.assert_not_called()


# ---------------------------------------------------------------------------
# _handle_new — adoption from the CLI's salt/batch/<jid>/new event
# ---------------------------------------------------------------------------


class TestHandleNew:
    def test_adopts_master_driven_batch(self, manager, opts):
        _write_state(opts, ["m1", "m2"], "JID1")
        manager._handle_new("JID1")
        assert "JID1" in manager.active_batches
        assert salt.utils.batch_state.read_active_index(opts) == {"JID1"}

    def test_ignores_cli_driven_batch(self, manager, opts):
        _write_state(opts, ["m1"], "JID-CLI", driver="cli")
        manager._handle_new("JID-CLI")
        assert "JID-CLI" not in manager.active_batches

    def test_handles_missing_batch_file(self, manager, opts):
        manager._handle_new("NONEXISTENT")
        assert manager.active_batches == set()

    def test_adoption_is_idempotent(self, manager, opts):
        _write_state(opts, ["m1"], "JID1")
        manager._handle_new("JID1")
        manager._handle_new("JID1")
        assert manager.active_batches == {"JID1"}


# ---------------------------------------------------------------------------
# _progress_one — the core step driven by returns and ticks
# ---------------------------------------------------------------------------


class TestProgressOne:
    def test_return_triggers_next_dispatch(self, manager, opts):
        _write_state(opts, ["m1", "m2", "m3", "m4"], "JID1", batch_size=2)
        manager.active_batches.add("JID1")
        # First progress pass — no returns yet, fills initial 2 slots.
        manager._progress_one("JID1", {}, now=1.0)
        manager.local.run_job.assert_called_once()
        first_args = manager.local.run_job.call_args.kwargs
        assert sorted(first_args["tgt"]) == ["m1", "m2"]
        assert first_args["jid"] == "JID1"

        # m1 and m2 return → state machine should publish m3, m4.
        manager.local.run_job.reset_mock()
        manager._progress_one(
            "JID1",
            {
                "m1": {"ret": True, "retcode": 0},
                "m2": {"ret": True, "retcode": 0},
            },
            now=2.0,
        )
        manager.local.run_job.assert_called_once()
        second_args = manager.local.run_job.call_args.kwargs
        assert sorted(second_args["tgt"]) == ["m3", "m4"]

    def test_completion_fires_batch_done_and_retires(self, manager, opts):
        _write_state(opts, ["m1", "m2"], "JID1", batch_size=2)
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)

        manager._progress_one("JID1", {}, now=1.0)  # dispatch
        manager._progress_one(
            "JID1",
            {
                "m1": {"ret": True, "retcode": 0},
                "m2": {"ret": True, "retcode": 0},
            },
            now=2.0,
        )
        manager.output.on_batch_done.assert_called_once()
        assert "JID1" not in manager.active_batches
        assert salt.utils.batch_state.read_active_index(opts) == set()

    def test_failhard_halts_and_fires_halted_event(self, manager, opts):
        _write_state(opts, ["m1", "m2", "m3"], "JID1", batch_size=2, failhard=True)
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)

        manager._progress_one("JID1", {}, now=1.0)
        manager._progress_one("JID1", {"m1": {"ret": False, "retcode": 1}}, now=2.0)

        manager.output.on_batch_done.assert_called_once()
        final_state = manager.output.on_batch_done.call_args.args[0]
        assert final_state["halted"] is True
        assert final_state["halted_reason"] == "failhard"
        assert "JID1" not in manager.active_batches

    def test_missing_batch_file_retires(self, manager, opts):
        manager.active_batches.add("GONE")
        salt.utils.batch_state.add_to_active_index("GONE", opts)
        manager._progress_one("GONE", {})
        assert "GONE" not in manager.active_batches
        assert salt.utils.batch_state.read_active_index(opts) == set()

    def test_halted_state_is_retired_without_dispatch(self, manager, opts):
        _write_state(
            opts,
            ["m1", "m2"],
            "JID1",
            batch_size=2,
            halted=True,
            halted_reason="stop",
        )
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)
        manager._progress_one("JID1", {})
        manager.local.run_job.assert_not_called()
        assert "JID1" not in manager.active_batches


# ---------------------------------------------------------------------------
# _handle_stop — graceful halt from the batch.stop runner
# ---------------------------------------------------------------------------


class TestHandleStop:
    def test_stop_sets_halted_and_fires_event(self, manager, opts):
        _write_state(opts, ["m1", "m2"], "JID1", batch_size=2)
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)

        manager._handle_stop("JID1", {"reason": "operator-requested"})

        state = salt.utils.batch_state.read_batch_state("JID1", opts)
        assert state["halted"] is True
        assert state["halted_reason"] == "operator-requested"
        manager.output.on_batch_halted.assert_called_once()
        assert "JID1" not in manager.active_batches

    def test_stop_is_idempotent_on_already_halted(self, manager, opts):
        _write_state(
            opts, ["m1"], "JID1", batch_size=1, halted=True, halted_reason="stop"
        )
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)

        manager._handle_stop("JID1", {"reason": "stop"})

        # No new halted event — already halted.
        manager.output.on_batch_halted.assert_not_called()
        assert "JID1" not in manager.active_batches

    def test_stop_default_reason_is_stop(self, manager, opts):
        _write_state(opts, ["m1"], "JID1", batch_size=1)
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)
        manager._handle_stop("JID1", {})
        state = salt.utils.batch_state.read_batch_state("JID1", opts)
        assert state["halted_reason"] == "stop"


# ---------------------------------------------------------------------------
# _handle_recover — Maintenance re-adoption path
# ---------------------------------------------------------------------------


class TestHandleRecover:
    def test_recover_readopts_and_ticks(self, manager, opts):
        _write_state(opts, ["m1", "m2"], "JID1", batch_size=2)
        # Manager has no memory of this batch (simulate crash/restart).
        assert "JID1" not in manager.active_batches
        manager._handle_recover("JID1")
        assert "JID1" in manager.active_batches
        # Recover should have kicked a progress pass, dispatching m1/m2.
        manager.local.run_job.assert_called_once()

    def test_recover_on_halted_batch_is_noop(self, manager, opts):
        _write_state(
            opts, ["m1"], "JID1", batch_size=1, halted=True, halted_reason="stop"
        )
        manager._handle_recover("JID1")
        assert "JID1" not in manager.active_batches
        manager.local.run_job.assert_not_called()

    def test_recover_on_cli_driven_batch_is_noop(self, manager, opts):
        _write_state(opts, ["m1"], "JID-CLI", batch_size=1, driver="cli")
        manager._handle_recover("JID-CLI")
        assert "JID-CLI" not in manager.active_batches


# ---------------------------------------------------------------------------
# _tick — housekeeping drives timeout and batch_wait expiry
# ---------------------------------------------------------------------------


class TestTick:
    def test_tick_progresses_every_active_batch(self, manager, opts):
        _write_state(opts, ["m1", "m2"], "JID-A", batch_size=2)
        _write_state(opts, ["x1", "x2"], "JID-B", batch_size=2)
        manager.active_batches.update(["JID-A", "JID-B"])

        manager._tick(now=1.0)

        # Both batches should have received their initial dispatch.
        assert manager.local.run_job.call_count == 2

    def test_tick_adopts_batches_from_index_added_externally(self, manager, opts):
        # Simulates the CLI having added the batch and indexed it while
        # the manager's salt/batch/<jid>/new event handling missed the
        # signal (or the manager was restarting).  The reconciliation
        # step in _tick must close this gap.
        _write_state(opts, ["m1", "m2"], "JID1", batch_size=2)
        salt.utils.batch_state.add_to_active_index("JID1", opts)
        assert "JID1" not in manager.active_batches
        manager._tick(now=1.0)
        assert "JID1" in manager.active_batches
        manager.local.run_job.assert_called_once()

    def test_tick_drives_timeout_detection(self, manager, opts):
        # Two minions dispatched at t=0; tick at t=1000 is way past the
        # timeout + gather_job_timeout window (70s default).
        _write_state(
            opts,
            ["m1", "m2"],
            "JID1",
            batch_size=2,
            active={"m1": 0.0, "m2": 0.0},
            pending=[],
        )
        manager.active_batches.add("JID1")
        salt.utils.batch_state.add_to_active_index("JID1", opts)

        manager._tick(now=1000.0)

        # Both timeouts → batch drains → on_batch_done fires → retired.
        manager.output.on_batch_done.assert_called_once()
        state = manager.output.on_batch_done.call_args.args[0]
        assert set(state["failed"].keys()) == {"m1", "m2"}
        assert state["failed"]["m1"] == "timeout"
        assert "JID1" not in manager.active_batches



"""
Unit tests for :mod:`salt.utils.batch_output`.

Pins the event tag vocabulary, payload shapes, and adapter dispatch
behavior.  The tag names are user-facing API; a change to any of them
is a breaking change and should require a conscious test update.
"""

from salt.utils import batch_output
from tests.support.mock import MagicMock

# ---------------------------------------------------------------------------
# Tag vocabulary
# ---------------------------------------------------------------------------


class TestTagVocabulary:
    def test_new(self):
        assert batch_output.tag_new("JID1") == "salt/batch/JID1/new"

    def test_progress(self):
        assert batch_output.tag_progress("JID1") == "salt/batch/JID1/progress"

    def test_complete(self):
        assert batch_output.tag_complete("JID1") == "salt/batch/JID1/complete"

    def test_halted(self):
        assert batch_output.tag_halted("JID1") == "salt/batch/JID1/halted"

    def test_recover(self):
        assert batch_output.tag_recover("JID1") == "salt/batch/JID1/recover"


# ---------------------------------------------------------------------------
# Payload builders
# ---------------------------------------------------------------------------


def _sample_state():
    return {
        "jid": "JID1",
        "fun": "test.ping",
        "tgt": "*",
        "tgt_type": "glob",
        "user": "alice",
        "driver": "master",
        "all_minions": ["m1", "m2", "m3", "m4"],
        "pending": ["m3", "m4"],
        "active": {"m2": 100.0},
        "done": {"m1": {"ret": True, "retcode": 0}},
        "failed": {},
        "batch_size": 2,
        "created": 90.0,
        "last_progress": 101.0,
        "halted": False,
        "halted_reason": None,
    }


class TestPayloads:
    def test_new_payload(self):
        payload = batch_output.new_payload(_sample_state())
        assert payload["jid"] == "JID1"
        assert payload["total_minions"] == 4
        assert payload["batch_size"] == 2
        assert payload["driver"] == "master"
        assert payload["created"] == 90.0

    def test_progress_payload(self):
        payload = batch_output.progress_payload(_sample_state(), iteration=3)
        assert payload["total"] == 4
        assert payload["completed"] == 1
        assert payload["active"] == 1
        assert payload["pending"] == 2
        assert payload["failed"] == 0
        assert payload["iter"] == 3
        assert payload["last_progress"] == 101.0

    def test_complete_payload(self):
        state = _sample_state()
        payload = batch_output.complete_payload(state, now=200.0)
        assert payload["completed"] == 1
        assert payload["failed"] == 0
        assert payload["total_minions"] == 4
        assert payload["duration"] == 110.0

    def test_halted_payload(self):
        state = _sample_state()
        state["halted"] = True
        state["halted_reason"] = "failhard"
        payload = batch_output.halted_payload(state, now=200.0)
        assert payload["reason"] == "failhard"
        assert payload["abandoned_active"] == ["m2"]
        assert payload["abandoned_pending"] == ["m3", "m4"]
        assert payload["duration"] == 110.0

    def test_recover_payload(self):
        payload = batch_output.recover_payload(_sample_state(), age_seconds=120.5)
        assert payload["reason"] == "stale"
        assert payload["age_seconds"] == 120.5
        assert payload["driver"] == "master"


# ---------------------------------------------------------------------------
# EventOutput dispatch
# ---------------------------------------------------------------------------


class TestEventOutput:
    def _make(self):
        event = MagicMock()
        return batch_output.EventOutput(opts={}, event=event), event

    def test_on_batch_new_fires_new_tag(self):
        adapter, event = self._make()
        state = _sample_state()
        adapter.on_batch_new(state)
        event.fire_event.assert_called_once()
        _, kwargs = event.fire_event.call_args
        data, tag = event.fire_event.call_args.args
        assert tag == "salt/batch/JID1/new"
        assert data["total_minions"] == 4

    def test_on_batch_done_normal_fires_complete(self):
        adapter, event = self._make()
        adapter.on_batch_done(_sample_state(), now=200.0)
        data, tag = event.fire_event.call_args.args
        assert tag == "salt/batch/JID1/complete"

    def test_on_batch_done_halted_fires_halted(self):
        adapter, event = self._make()
        state = _sample_state()
        state["halted"] = True
        state["halted_reason"] = "failhard"
        adapter.on_batch_done(state, now=200.0)
        data, tag = event.fire_event.call_args.args
        assert tag == "salt/batch/JID1/halted"
        assert data["reason"] == "failhard"

    def test_on_batch_progress_fires_progress_tag(self):
        adapter, event = self._make()
        adapter.on_batch_progress(_sample_state(), iteration=5)
        data, tag = event.fire_event.call_args.args
        assert tag == "salt/batch/JID1/progress"
        assert data["iter"] == 5


class TestSilentOutput:
    def test_all_hooks_are_noops(self):
        adapter = batch_output.SilentOutput()
        state = _sample_state()
        # None of these should raise, regardless of what's passed.
        adapter.on_batch_new(state)
        adapter.on_batch_progress(state)
        adapter.on_batch_complete(state)
        adapter.on_batch_halted(state)
        adapter.on_batch_recover(state, age_seconds=0)
        adapter.on_batch_start(["m1"])
        adapter.on_minion_return("m1", {})
        adapter.on_minion_timeout("m1")
        adapter.on_batch_done(state)



"""
Verify salt-ssh can use imported map files in states
when the map files are in another directory outside of
sls files importing them.
"""

import pytest

from tests.support.helpers import system_python_version

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
    pytest.mark.skipif(
        system_python_version() < (3, 10),
        reason="System python too old for these tests",
    ),
]


@pytest.mark.parametrize(
    "ssh_cmd",
    [
        "state.sls",
        "state.highstate",
        "state.apply",
        "state.show_top",
        "state.show_highstate",
        "state.show_low_sls",
        "state.show_lowstate",
        "state.sls_id",
        "state.show_sls",
        "state.top",
    ],
)
def test_state_with_import_dir(state_tree_dir, salt_ssh_cli_parameterized, ssh_cmd):
    # Note: Removed -w and -t flags since salt_ssh_cli_parameterized handles deployment type
    if ssh_cmd in ("state.sls", "state.show_low_sls", "state.show_sls"):
        ret = salt_ssh_cli_parameterized.run(ssh_cmd, "testdir")
    elif ssh_cmd == "state.top":
        ret = salt_ssh_cli_parameterized.run(ssh_cmd, "top.sls")
    elif ssh_cmd == "state.sls_id":
        ret = salt_ssh_cli_parameterized.run(ssh_cmd, "Ok with def", "testdir")
    else:
        ret = salt_ssh_cli_parameterized.run(ssh_cmd)
    assert ret.returncode == 0
    if ssh_cmd == "state.show_top":
        assert ret.data == {"base": ["testdir", "master_tops_test"]} or {
            "base": ["testdir"]
        }
    elif ssh_cmd in ("state.show_highstate", "state.show_sls"):
        assert ret.data == {
            "Ok with def": {
                "__sls__": "testdir",
                "__env__": "base",
                "test": ["succeed_without_changes", {"order": 10000}],
            }
        }
    elif ssh_cmd in ("state.show_low_sls", "state.show_lowstate", "state.show_sls"):
        assert ret.data == [
            {
                "state": "test",
                "name": "Ok with def",
                "__sls__": "testdir",
                "__env__": "base",
                "__id__": "Ok with def",
                "order": 10000,
                "fun": "succeed_without_changes",
            }
        ]
    else:
        assert ret.data["test_|-Ok with def_|-Ok with def_|-succeed_without_changes"][
            "result"
        ]
    assert ret.data



"""
The Salt Key backend API and interface used by the CLI. The Key class can be
used to manage salt keys directly without interfacing with the CLI.
"""

import fnmatch
import itertools
import logging
import os
import sys

import salt.cache
import salt.client
import salt.crypt
import salt.exceptions
import salt.payload
import salt.transport
import salt.utils.args
import salt.utils.crypt
import salt.utils.data
import salt.utils.event
import salt.utils.files
import salt.utils.json
import salt.utils.kinds
import salt.utils.minions
import salt.utils.sdb
import salt.utils.stringutils
import salt.utils.user
from salt.utils.decorators import cached_property

log = logging.getLogger(__name__)


def get_key(opts):
    return Key(opts)


class KeyCLI:
    """
    Manage key CLI operations
    """

    CLI_KEY_MAP = {
        "list": "list_status",
        "delete": "delete_key",
        "gen_signature": "gen_keys_signature",
        "print": "key_str",
    }

    def __init__(self, opts):
        self.opts = opts
        import salt.wheel

        self.client = salt.wheel.WheelClient(opts)
        # instantiate the key object for masterless mode
        if not opts.get("eauth"):
            self.key = get_key(opts)
        else:
            self.key = Key

        self.auth = {}

    def _update_opts(self):
        # get the key command
        for cmd in (
            "gen_keys",
            "gen_signature",
            "list",
            "list_all",
            "print",
            "print_all",
            "accept",
            "accept_all",
            "reject",
            "reject_all",
            "delete",
            "delete_all",
            "finger",
            "finger_all",
            "list_all",
        ):  # last is default
            if self.opts[cmd]:
                break
        # set match if needed
        if not cmd.startswith("gen_"):
            if cmd == "list_all":
                self.opts["match"] = "all"
            elif cmd.endswith("_all"):
                self.opts["match"] = "*"
            else:
                self.opts["match"] = self.opts[cmd]
            if cmd.startswith("accept"):
                self.opts["include_rejected"] = (
                    self.opts["include_all"] or self.opts["include_rejected"]
                )
                self.opts["include_accepted"] = False
            elif cmd.startswith("reject"):
                self.opts["include_accepted"] = (
                    self.opts["include_all"] or self.opts["include_accepted"]
                )
                self.opts["include_rejected"] = False
        elif cmd == "gen_keys":
            self.opts["keydir"] = self.opts["gen_keys_dir"]
            self.opts["keyname"] = self.opts["gen_keys"]
        # match is set to opts, now we can forget about *_all commands
        self.opts["fun"] = cmd.replace("_all", "")

    def _init_auth(self):
        if self.auth:
            return

        low = {}
        skip_perm_errors = self.opts["eauth"] != ""

        if self.opts["eauth"]:
            if "token" in self.opts:
                try:
                    with salt.utils.files.fopen(
                        os.path.join(self.opts["cachedir"], ".root_key"), "r"
                    ) as fp_:
                        low["key"] = salt.utils.stringutils.to_unicode(fp_.readline())
                except OSError:
                    low["token"] = self.opts["token"]

            # If using eauth and a token hasn't already been loaded into
            # low, prompt the user to enter auth credentials
            if "token" not in low and "key" not in low and self.opts["eauth"]:
                # This is expensive. Don't do it unless we need to.
                import salt.auth

                resolver = salt.auth.Resolver(self.opts)
                res = resolver.cli(self.opts["eauth"])
                if self.opts["mktoken"] and res:
                    tok = resolver.token_cli(self.opts["eauth"], res)
                    if tok:
                        low["token"] = tok.get("token", "")
                if not res:
                    log.error("Authentication failed")
                    return {}
                low.update(res)
                low["eauth"] = self.opts["eauth"]
        else:
            # late import to avoid circular import
            import salt.utils.master

            low["user"] = salt.utils.user.get_specific_user()
            low["key"] = salt.utils.master.get_master_key(
                low["user"], self.opts, skip_perm_errors
            )

        self.auth = low

    def _get_args_kwargs(self, fun, args=None):
        argspec = salt.utils.args.get_function_argspec(fun)
        if args is None:
            args = []
            if argspec.args:
                # Iterate in reverse order to ensure we get the correct default
                # value for the positional argument.
                for arg, default in itertools.zip_longest(
                    reversed(argspec.args), reversed(argspec.defaults or ())
                ):
                    args.append(self.opts.get(arg, default))
            # Reverse the args so that they are in the correct order
            args = args[::-1]

        if argspec.keywords is None:
            kwargs = {}
        else:
            args, kwargs = salt.minion.load_args_and_kwargs(fun, args)
        return args, kwargs

    def _run_cmd(self, cmd, args=None):
        if not self.opts.get("eauth"):
            cmd = self.CLI_KEY_MAP.get(cmd, cmd)
            fun = getattr(self.key, cmd)
            args, kwargs = self._get_args_kwargs(fun, args)
            ret = fun(*args, **kwargs)
            if (
                isinstance(ret, dict)
                and "local" in ret
                and cmd not in ("finger", "finger_all")
            ):
                ret.pop("local", None)
            return ret

        if cmd in ("accept", "reject", "delete") and args is None:
            args = self.opts.get("match_dict", {}).get("minions")
        fstr = f"key.{cmd}"
        fun = self.client.functions[fstr]
        args, kwargs = self._get_args_kwargs(fun, args)

        low = {
            "fun": fstr,
            "arg": args,
            "kwarg": kwargs,
        }

        self._init_auth()
        low.update(self.auth)

        # Execute the key request!
        ret = self.client.cmd_sync(low)

        ret = ret["data"]["return"]
        if (
            isinstance(ret, dict)
            and "local" in ret
            and cmd not in ("finger", "finger_all")
        ):
            ret.pop("local", None)

        return ret

    def _filter_ret(self, cmd, ret):
        if cmd.startswith("delete"):
            return ret

        keys = {}
        if self.key.PEND in ret:
            keys[self.key.PEND] = ret[self.key.PEND]
        if self.opts["include_accepted"] and bool(ret.get(self.key.ACC)):
            keys[self.key.ACC] = ret[self.key.ACC]
        if self.opts["include_rejected"] and bool(ret.get(self.key.REJ)):
            keys[self.key.REJ] = ret[self.key.REJ]
        if self.opts["include_denied"] and bool(ret.get(self.key.DEN)):
            keys[self.key.DEN] = ret[self.key.DEN]
        return keys

    def _print_no_match(self, cmd, match):
        statuses = ["unaccepted"]
        if self.opts["include_accepted"]:
            statuses.append("accepted")
        if self.opts["include_rejected"]:
            statuses.append("rejected")
        if self.opts["include_denied"]:
            statuses.append("denied")
        if len(statuses) == 1:
            stat_str = statuses[0]
        else:
            stat_str = "{} or {}".format(", ".join(statuses[:-1]), statuses[-1])
        msg = f"The key glob '{match}' does not match any {stat_str} keys."
        print(msg)

    def run(self):
        """
        Run the logic for saltkey
        """
        self._update_opts()
        cmd = self.opts["fun"]

        veri = None
        ret = None
        try:
            if cmd in ("accept", "reject", "delete"):
                ret = self._run_cmd("glob_match")
                if not isinstance(ret, dict):
                    salt.output.display_output(ret, "key", opts=self.opts)
                    return ret
                ret = self._filter_ret(cmd, ret)
                if not ret:
                    self._print_no_match(cmd, self.opts["match"])
                    return
                print(
                    "The following keys are going to be {}ed:".format(cmd.rstrip("e"))
                )
                salt.output.display_output(ret, "key", opts=self.opts)

                if not self.opts.get("yes", False):
                    try:
                        if cmd.startswith("delete"):
                            veri = input("Proceed? [N/y] ")
                            if not veri:
                                veri = "n"
                        else:
                            veri = input("Proceed? [n/Y] ")
                            if not veri:
                                veri = "y"
                    except KeyboardInterrupt:
                        raise SystemExit("\nExiting on CTRL-c")
                # accept/reject/delete the same keys we're printed to the user
                self.opts["match_dict"] = ret
                self.opts.pop("match", None)
                list_ret = ret

            if veri is None or veri.lower().startswith("y"):
                ret = self._run_cmd(cmd)
                if cmd in ("accept", "reject", "delete"):
                    if cmd == "delete":
                        ret = list_ret
                    for minions in ret.values():
                        for minion in minions:
                            print(
                                "Key for minion {} {}ed.".format(
                                    minion, cmd.rstrip("e")
                                )
                            )
                elif isinstance(ret, dict):
                    salt.output.display_output(ret, "key", opts=self.opts)
                else:
                    salt.output.display_output({"return": ret}, "key", opts=self.opts)
        except salt.exceptions.SaltException as exc:
            ret = f"{exc}"
            if not self.opts.get("quiet", False):
                salt.output.display_output(ret, "nested", self.opts)
        except Exception as exc:  # pylint: disable=broad-except
            # dont swallow unexpected exceptions in salt-key
            log.exception(exc)

        return ret


class Key:
    """
    The object that encapsulates saltkey actions
    """

    ACC = "minions"
    PEND = "minions_pre"
    REJ = "minions_rejected"
    DEN = "minions_denied"

    # handle transitions from legacy naming to simpler new format
    STATE_MAP = {"accepted": ACC, "rejected": REJ, "pending": PEND, "denied": DEN}
    DIR_MAP = {v: k for k, v in STATE_MAP.items()}

    ACT_MAP = {
        ACC: "accept",
        REJ: "reject",
        PEND: "pend",
        DEN: "denied",
    }

    def __init__(self, opts, io_loop=None):
        self.opts = opts
        self.cache = salt.cache.Cache(opts, driver=self.opts["keys.cache_driver"])
        if self.opts.get("cluster_id", None) is not None:
            self.pki_dir = self.opts.get("cluster_pki_dir", "")
        else:
            self.pki_dir = self.opts.get("pki_dir", "")
        self._kind = self.opts.get("__role", "")  # application kind
        if self._kind not in salt.utils.kinds.APPL_KINDS:
            emsg = f"Invalid application kind = '{self._kind}'."
            log.error(emsg)
            raise ValueError(emsg)
        self.passphrase = salt.utils.sdb.sdb_get(
            self.opts.get("signing_key_pass"), self.opts
        )
        self.io_loop = io_loop

    @cached_property
    def master_keys(self):
        return salt.crypt.MasterKeys(self.opts)

    @cached_property
    def event(self):
        return salt.utils.event.get_event(
            self._kind,
            self.opts["sock_dir"],
            opts=self.opts,
            listen=False,
            io_loop=self.io_loop,
        )

    def _check_minions_directories(self):
        """
        Return the minion keys directory paths
        """
        minions_accepted = os.path.join(self.pki_dir, self.ACC)
        minions_pre = os.path.join(self.pki_dir, self.PEND)
        minions_rejected = os.path.join(self.pki_dir, self.REJ)

        minions_denied = os.path.join(self.pki_dir, self.DEN)
        return minions_accepted, minions_pre, minions_rejected, minions_denied

    def _get_key_attrs(self, keydir, keyname, keysize, user):
        cache = None
        if not keydir:
            if "gen_keys_dir" in self.opts:
                keydir = self.opts["gen_keys_dir"]
            else:
                keydir = self.pki_dir
        cache = salt.cache.Cache(
            self.opts, driver=self.opts["keys.cache_driver"], cachedir=keydir, user=user
        )
        if not keyname:
            if "gen_keys" in self.opts:
                keyname = self.pki_dir
            else:
                keyname = "minion"
        if not keysize:
            keysize = self.opts["keysize"]
        return keydir, keyname, keysize, user, cache

    def gen_keys(self, keydir=None, keyname=None, keysize=None, user=None):
        """
        Generate minion RSA public keypair
        """
        keydir, keyname, keysize, user, cache = self._get_key_attrs(
            keydir, keyname, keysize, user
        )
        priv = self.master_keys.find_or_create_keys(
            keyname, keysize=keysize, cache=cache
        )
        return salt.utils.crypt.pem_finger(key=priv.public_key())

    def gen_keys_signature(
        self, priv, pub, signature_path, auto_create=False, keysize=None
    ):
        """
        Generate master public-key-signature
        """
        # check given pub-key
        if pub:
            if not os.path.isfile(pub):
                return f"Public-key {pub} does not exist"
        # default to master.pub
        else:
            mpub = self.pki_dir + "/" + "master.pub"
            if os.path.isfile(mpub):
                pub = mpub

        # check given priv-key
        if priv:
            if not os.path.isfile(priv):
                return f"Private-key {priv} does not exist"
        # default to master_sign.pem
        else:
            mpriv = self.pki_dir + "/" + "master_sign.pem"
            if os.path.isfile(mpriv):
                priv = mpriv

        if priv:
            priv = salt.crypt.PrivateKey.from_file(priv)
        else:
            if auto_create:
                log.debug(
                    "Generating new signing key-pair .%s.* in %s",
                    self.opts["master_sign_key_name"],
                    self.pki_dir,
                )
                # we force re-create as master_keys init also does the same
                # creation without these kwarg overrides
                priv = self.master_keys.sign_key = self.master_keys.find_or_create_keys(
                    name=self.opts["master_sign_key_name"],
                    keysize=keysize or self.opts["keysize"],
                    passphrase=self.passphrase,
                    force=True,
                )
            else:
                return "No usable private-key found"

        if pub:
            pub = salt.crypt.PublicKey.from_file(pub).key
        else:
            return "No usable public-key found"

        log.debug("Using public-key %s", pub)
        log.debug("Using private-key %s", priv)

        if signature_path:
            if not os.path.isdir(signature_path):
                log.debug("target directory %s does not exist", signature_path)
            sign_path = signature_path + "/" + self.master_keys.master_pubkey_signature
        else:
            sign_path = None

        return self.master_keys.gen_signature(priv, pub, sign_path)

    def check_minion_cache(self, preserve_minions=None):
        """
        Check the minion cache to make sure that old minion data is cleared

        Optionally, pass in a list of minions which should have their caches
        preserved. To preserve all caches, set __opts__['preserve_minion_cache']
        """
        if self.opts.get("preserve_minion_cache", False):
            return

        if preserve_minions is None:
            preserve_minions = []
        preserve_minions = set(preserve_minions)

        keys = self.list_keys()

        for val in keys.values():
            preserve_minions.update(val)

        # we use a new cache instance here as we dont want the key cache
        cache = salt.cache.factory(self.opts)

        for bank in ["grains", "pillar"]:
            clist = set(cache.list(bank))
            for minion in clist - preserve_minions:
                # pillar optionally encodes pillarenv in the key as minion:$pillarenv
                if ":" in minion and minion.split(":")[0] in preserve_minions:
                    continue
                cache.flush(bank, minion)

    def check_master(self):
        """
        Log if the master is not running

        :rtype: bool
        :return: Whether or not the master is running
        """
        if not os.path.exists(os.path.join(self.opts["sock_dir"], "publish_pull.ipc")):
            return False
        return True

    def glob_match(self, match, full=False):
        """
        Accept a glob which to match the of a key and return the key's location
        """
        if full:
            matches = self.all_keys()
        else:
            matches = self.list_keys()
        ret = {}
        if "," in match and isinstance(match, str):
            match = match.split(",")
        if not isinstance(match, list):
            match = [match]
        for status, keys in matches.items():
            if match == ["*"] and keys:
                ret[status] = keys
                continue
            for key in salt.utils.data.sorted_ignorecase(keys):
                for match_item in match:
                    if fnmatch.fnmatch(key, match_item):
                        if status not in ret:
                            ret[status] = []
                        ret[status].append(key)
        return ret

    def list_match(self, match):
        """
        Accept a glob which to match the of a key and return the key's location
        """
        ret = {}
        if isinstance(match, str):
            match = match.split(",")

        for name in match:
            key = self.cache.fetch("keys", name)
            if key:
                try:
                    ret.setdefault(self.STATE_MAP[key["state"]], [])
                    ret[self.STATE_MAP[key["state"]]].append(name)
                except KeyError:
                    log.error("unexpected key state returned for %s: %s", name, key)

            denied_keys = self.cache.fetch("denied_keys", name)
            if denied_keys:
                ret.setdefault(self.DEN, [])
                ret[self.DEN].append(name)
        return ret

    def dict_match(self, match_dict):
        """
        Accept a dictionary of keys and return the current state of the
        specified keys
        """
        ret = {}
        cur_keys = self.list_keys()
        for status, keys in match_dict.items():
            for key in salt.utils.data.sorted_ignorecase(keys):
                for keydir in (self.ACC, self.PEND, self.REJ, self.DEN):
                    if keydir and fnmatch.filter(cur_keys.get(keydir, []), key):
                        ret.setdefault(keydir, []).append(key)
        return ret

    def list_keys(self):
        """
        Return a dict of managed keys and what the key status are
        """
        if self.opts.get("key_cache") == "sched":
            acc = "accepted"

            cache_file = os.path.join(self.opts["pki_dir"], acc, ".key_cache")
            if self.opts["key_cache"] and os.path.exists(cache_file):
                log.debug("Returning cached minion list")
                with salt.utils.files.fopen(cache_file, mode="rb") as fn_:
                    return salt.payload.load(fn_)

        # Use cache layer's optimized bulk fetch
        if self.opts.get("pki_index_enabled") is True:
            from salt.utils import (
                pki as pki_utils,  # pylint: disable=import-outside-toplevel
            )

            index = pki_utils.PkiIndex(self.opts)
            items = index.list_items()
            if items:
                ret = {
                    "minions_pre": [],
                    "minions_rejected": [],
                    "minions": [],
                    "minions_denied": [],
                }
                for id_, state in items:
                    if state == "accepted":
                        ret["minions"].append(id_)
                    elif state == "pending":
                        ret["minions_pre"].append(id_)
                    elif state == "rejected":
                        ret["minions_rejected"].append(id_)

                # Sort for consistent CLI output
                for key in ret:
                    ret[key] = salt.utils.data.sorted_ignorecase(ret[key])

                # Denied keys are not in the index currently
                for id_ in salt.utils.data.sorted_ignorecase(
                    self.cache.list("denied_keys")
                ):
                    ret["minions_denied"].append(id_)
                return ret

        ret = {
            "minions_pre": [],
            "minions_rejected": [],
            "minions": [],
            "minions_denied": [],
        }
        for id_ in salt.utils.data.sorted_ignorecase(self.cache.list("keys")):
            key = self.cache.fetch("keys", id_)

            if key["state"] == "accepted":
                ret["minions"].append(id_)
            elif key["state"] == "pending":
                ret["minions_pre"].append(id_)
            elif key["state"] == "rejected":
                ret["minions_rejected"].append(id_)

        for id_ in salt.utils.data.sorted_ignorecase(self.cache.list("denied_keys")):
            ret["minions_denied"].append(id_)
        return ret

    def local_keys(self):
        """
        Return a dict of local keys
        """
        ret = {"local": []}
        for key in salt.utils.data.sorted_ignorecase(self.cache.list("master_keys")):
            if key.endswith(".pub") or key.endswith(".pem"):
                ret["local"].append(key)
        return ret

    def all_keys(self):
        """
        Merge managed keys with local keys
        """
        keys = self.list_keys()
        keys.update(self.local_keys())
        return keys

    def list_status(self, match):
        """
        Return a dict of managed keys under a named status
        """
        ret = self.all_keys()
        if match.startswith("acc"):
            return {
                "minions": salt.utils.data.sorted_ignorecase(ret.get("minions", []))
            }
        elif match.startswith("pre") or match.startswith("un"):
            return {
                "minions_pre": salt.utils.data.sorted_ignorecase(
                    ret.get("minions_pre", [])
                )
            }
        elif match.startswith("rej"):
            return {
                "minions_rejected": salt.utils.data.sorted_ignorecase(
                    ret.get("minions_rejected", [])
                )
            }
        elif match.startswith("den"):
            return {
                "minions_denied": salt.utils.data.sorted_ignorecase(
                    ret.get("minions_denied", [])
                )
            }
        elif match.startswith("all"):
            return ret
        # this should never be reached
        return {}

    def key_str(self, match):
        """
        Return the specified public key or keys based on a glob
        """
        ret = {}
        for status, keys in self.glob_match(match).items():
            ret[status] = {}
            for key in salt.utils.data.sorted_ignorecase(keys):
                if status == self.DEN:
                    denied = self.cache.fetch("denied_keys", key)
                    if len(denied) == 1:
                        ret[status][key] = denied[0]
                    else:
                        ret[status][key] = denied
                else:
                    ret[status][key] = self.cache.fetch("keys", key).get("pub")
        return ret

    def key_str_all(self):
        """
        Return all managed key strings
        """
        return self.key_str("*")

    def change_state(
        self,
        from_state,
        to_state,
        match=None,
        match_dict=None,
        include_rejected=False,
        include_denied=False,
        include_accepted=False,
    ):
        """
        change key state from one state to another
        """
        if match is not None:
            matches = self.glob_match(match)
        elif match_dict is not None and isinstance(match_dict, dict):
            matches = match_dict
        else:
            matches = {}
        keydirs = [from_state]
        if include_rejected:
            keydirs.append(self.REJ)
        if include_denied:
            keydirs.append(self.DEN)
        if include_accepted:
            keydirs.append(self.ACC)

        invalid_keys = []
        for keydir in keydirs:
            for keyname in matches.get(keydir, []):
                if to_state == self.DEN:
                    key = self.cache.fetch("keys", keyname)
                    self.cache.flush("keys", keyname)
                    self.cache.store("denied_keys", keyname, [key["pub"]])
                else:
                    if keydir == self.DEN:
                        # denied keys can be many per id, but we assume first for legacy
                        pub = self.cache.fetch("denied_keys", keyname)[0]
                        self.cache.flush("denied_keys", keyname)
                        key = {"pub": pub}
                    else:
                        key = self.cache.fetch("keys", keyname)

                    try:
                        salt.crypt.PublicKey.from_str(key["pub"])
                    except salt.exceptions.InvalidKeyError:
                        log.error("Invalid RSA public key: %s", keyname)
                        invalid_keys.append(keyname)
                        continue

                    key["state"] = self.DIR_MAP[to_state]
                    self.cache.store("keys", keyname, key)

                eload = {"result": True, "act": self.DIR_MAP[to_state], "id": keyname}
                self.event.fire_event(eload, salt.utils.event.tagify(prefix="key"))

        for key in invalid_keys:
            sys.stderr.write(f"Unable to accept invalid key for {key}.\n")

        return self.glob_match(match) if match is not None else self.dict_match(matches)

    def accept(
        self, match=None, match_dict=None, include_rejected=False, include_denied=False
    ):
        """
        Accept public keys. If "match" is passed, it is evaluated as a glob.
        Pre-gathered matches can also be passed via "match_dict".
        """
        return self.change_state(
            self.PEND,
            self.ACC,
            match,
            match_dict,
            include_rejected=include_rejected,
            include_denied=include_denied,
        )

    def accept_all(self):
        """
        Accept all keys in pre
        """
        return self.accept(match="*")

    def delete_key(
        self, match=None, match_dict=None, preserve_minions=None, revoke_auth=False
    ):
        """
        Delete public keys. If "match" is passed, it is evaluated as a glob.
        Pre-gathered matches can also be passed via "match_dict".

        To preserve the master caches of minions who are matched, set preserve_minions
        """
        if match is not None:
            matches = self.glob_match(match)
        elif match_dict is not None and isinstance(match_dict, dict):
            matches = match_dict
        else:
            matches = {}
        with salt.client.get_local_client(mopts=self.opts) as client:
            for status, keys in matches.items():
                for key in keys:
                    try:
                        if revoke_auth:
                            if self.opts.get("rotate_aes_key") is False:
                                print(
                                    "Immediate auth revocation specified but AES key"
                                    " rotation not allowed. Minion will not be"
                                    " disconnected until the master AES key is rotated."
                                )
                            else:
                                try:
                                    client.cmd_async(key, "saltutil.revoke_auth")
                                except salt.exceptions.SaltClientError:
                                    print(
                                        "Cannot contact Salt master. "
                                        "Connection for {} will remain up until "
                                        "master AES key is rotated or auth is revoked "
                                        "with 'saltutil.revoke_auth'.".format(key)
                                    )
                        if status == "minions_denied":
                            self.cache.flush("denied_keys", key)
                        else:
                            self.cache.flush("keys", key)
                        eload = {"result": True, "act": "delete", "id": key}
                        self.event.fire_event(
                            eload, salt.utils.event.tagify(prefix="key")
                        )
                    except OSError:
                        pass
        if self.opts.get("preserve_minions") is True:
            self.check_minion_cache(preserve_minions=matches.get("minions", []))
        else:
            self.check_minion_cache()
        if self.opts.get("rotate_aes_key"):
            salt.crypt.dropfile(
                self.opts["cachedir"], self.opts["user"], self.opts["id"]
            )

        return self.glob_match(match) if match is not None else self.dict_match(matches)

    def delete_den(self):
        """
        Delete all denied keys
        """
        self.cache.flush("denied_keys")
        self.check_minion_cache()
        return self.list_keys()

    def delete_all(self):
        """
        Delete all keys
        """
        for status, keys in self.list_keys().items():
            for key in keys:
                try:
                    self.cache.flush("keys", key)
                    eload = {"result": True, "act": "delete", "id": key}
                    self.event.fire_event(eload, salt.utils.event.tagify(prefix="key"))
                except OSError:
                    pass
        self.check_minion_cache()
        if self.opts.get("rotate_aes_key"):
            salt.crypt.dropfile(
                self.opts["cachedir"], self.opts["user"], self.opts["id"]
            )
        return self.list_keys()

    def reject(
        self, match=None, match_dict=None, include_accepted=False, include_denied=False
    ):
        """
        Reject public keys. If "match" is passed, it is evaluated as a glob.
        Pre-gathered matches can also be passed via "match_dict".
        """
        ret = self.change_state(
            self.PEND,
            self.REJ,
            match,
            match_dict,
            include_accepted=include_accepted,
            include_denied=include_denied,
        )
        self.check_minion_cache()
        if self.opts.get("rotate_aes_key"):
            salt.crypt.dropfile(
                self.opts["cachedir"], self.opts["user"], self.opts["id"]
            )
        return ret

    def reject_all(self):
        """
        Reject all keys in pre
        """
        self.reject(match="*")
        self.check_minion_cache()
        if self.opts.get("rotate_aes_key"):
            salt.crypt.dropfile(
                self.opts["cachedir"], self.opts["user"], self.opts["id"]
            )
        return self.list_keys()

    def finger(self, match, hash_type=None):
        """
        Return the fingerprint for a specified key
        """
        if hash_type is None:
            hash_type = self.opts["hash_type"]

        matches = self.glob_match(match, full=True)
        ret = {}
        for status, keys in matches.items():
            ret[status] = {}
            for key in keys:
                if status == "minions_denied":
                    denied = self.cache.fetch("denied_keys", key)
                    for den in denied:
                        finger = salt.utils.crypt.pem_finger(
                            key=den.encode("utf-8"), sum_type=hash_type
                        )
                        ret[status].setdefault(key, []).append(finger)
                    # brush over some dumb backcompat with how denied keys work
                    # with the legacy system
                    if len(denied) == 1:
                        ret[status][key] = ret[status][key][0]
                else:
                    if status == "local":
                        pub = self.cache.fetch("master_keys", key).encode("utf-8")
                    else:
                        pub = self.cache.fetch("keys", key)["pub"].encode("utf-8")
                    ret[status][key] = salt.utils.crypt.pem_finger(
                        key=pub, sum_type=hash_type
                    )
        return ret

    def finger_all(self, hash_type=None):
        """
        Return fingerprints for all keys
        """
        if hash_type is None:
            hash_type = self.opts["hash_type"]

        return self.finger("*", hash_type=hash_type)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.event.destroy()



"""
Direct unit tests for the non-``progress_batch`` helpers in
``salt.utils.batch_state``.

The conformance scenarios in :mod:`.test_conformance` cover the state
machine itself.  These tests pin the surrounding helpers:
``get_batch_size``, ``create_batch_state``, ``is_batch_done``, and the
``.batch.p`` / ``batch_active.p`` persistence functions.
"""

import pytest

import salt.exceptions
from salt.utils.batch_state import (
    add_to_active_index,
    create_batch_state,
    get_batch_size,
    is_batch_done,
    read_active_index,
    read_batch_state,
    remove_from_active_index,
    write_active_index,
    write_batch_state,
)

# ---------------------------------------------------------------------------
# get_batch_size
# ---------------------------------------------------------------------------


class TestGetBatchSize:
    def test_integer_spec(self):
        assert get_batch_size(10, 100) == 10

    def test_integer_string_spec(self):
        assert get_batch_size("10", 100) == 10

    def test_percentage_25(self):
        assert get_batch_size("25%", 100) == 25

    def test_percentage_100_exact(self):
        assert get_batch_size("100%", 5) == 5

    def test_percentage_rounds_up_below_one(self):
        # 1% of 50 = 0.5, must round up to 1.
        assert get_batch_size("1%", 50) == 1

    def test_percentage_fractional_truncates_when_above_one(self):
        # 33% of 10 = 3.3, sync batch floors via int().
        assert get_batch_size("33%", 10) == 3

    def test_zero_batch_floors_to_one(self):
        # Zero would deadlock the state machine; floor to 1.
        assert get_batch_size(0, 100) == 1

    def test_negative_batch_floors_to_one(self):
        assert get_batch_size(-5, 100) == 1

    def test_empty_minion_list_returns_at_least_one(self):
        assert get_batch_size("50%", 0) == 1

    def test_invalid_string_raises_invocation_error(self):
        with pytest.raises(salt.exceptions.SaltInvocationError):
            get_batch_size("not-a-number", 10)

    def test_none_raises_invocation_error(self):
        with pytest.raises(salt.exceptions.SaltInvocationError):
            get_batch_size(None, 10)


# ---------------------------------------------------------------------------
# create_batch_state
# ---------------------------------------------------------------------------


class TestCreateBatchState:
    def test_populates_required_fields(self):
        opts = {
            "batch": 2,
            "fun": "test.ping",
            "arg": [],
            "tgt": "*",
            "tgt_type": "glob",
            "timeout": 30,
            "gather_job_timeout": 5,
        }
        state = create_batch_state(opts, ["m1", "m2", "m3"], "JID1", now=100.0)
        assert state["jid"] == "JID1"
        assert state["all_minions"] == ["m1", "m2", "m3"]
        assert state["pending"] == ["m1", "m2", "m3"]
        assert state["active"] == {}
        assert state["done"] == {}
        assert state["failed"] == {}
        assert state["wait"] == []
        assert state["batch_size"] == 2
        assert state["fun"] == "test.ping"
        assert state["tgt"] == "*"
        assert state["tgt_type"] == "glob"
        assert state["timeout"] == 30
        assert state["gather_job_timeout"] == 5
        assert state["created"] == 100.0
        assert state["last_progress"] == 100.0
        assert state["halted"] is False
        assert state["halted_reason"] is None
        assert state["driver"] == "cli"

    def test_percentage_batch_spec(self):
        state = create_batch_state(
            {"batch": "50%"}, ["m1", "m2", "m3", "m4"], "JID", now=1.0
        )
        assert state["batch_size"] == 2

    def test_driver_defaults_cli_but_can_be_master(self):
        state = create_batch_state(
            {"batch": 1}, ["m1"], "JID", driver="master", now=1.0
        )
        assert state["driver"] == "master"

    def test_pending_is_copy_not_alias(self):
        # Mutating all_minions later must not bleed into pending.
        minions = ["m1", "m2"]
        state = create_batch_state({"batch": 1}, minions, "JID", now=1.0)
        minions.append("m3")
        assert state["pending"] == ["m1", "m2"]
        assert state["all_minions"] == ["m1", "m2"]

    def test_ret_falls_back_to_return(self):
        state = create_batch_state(
            {"batch": 1, "return": "slack"}, ["m1"], "JID", now=1.0
        )
        assert state["ret"] == "slack"

    def test_user_defaults_to_root(self):
        state = create_batch_state({"batch": 1}, ["m1"], "JID", now=1.0)
        assert state["user"] == "root"


# ---------------------------------------------------------------------------
# is_batch_done
# ---------------------------------------------------------------------------


class TestIsBatchDone:
    def test_drained_state_is_done(self):
        state = {"pending": [], "active": {}, "halted": False}
        assert is_batch_done(state) is True

    def test_halted_state_is_done(self):
        state = {"pending": ["m1"], "active": {"m2": 1}, "halted": True}
        assert is_batch_done(state) is True

    def test_pending_prevents_done(self):
        state = {"pending": ["m1"], "active": {}, "halted": False}
        assert is_batch_done(state) is False

    def test_active_prevents_done(self):
        state = {"pending": [], "active": {"m1": 1.0}, "halted": False}
        assert is_batch_done(state) is False


# ---------------------------------------------------------------------------
# Persistence helpers — .batch.p and batch_active.p
# ---------------------------------------------------------------------------


@pytest.fixture
def opts(tmp_path):
    return {
        "cachedir": str(tmp_path),
        "hash_type": "sha256",
    }


class TestBatchStatePersistence:
    def test_round_trip(self, opts):
        state = {
            "jid": "JID1",
            "pending": ["m1"],
            "done": {"m2": True},
            "halted": False,
        }
        write_batch_state("JID1", state, opts)
        assert read_batch_state("JID1", opts) == state

    def test_read_missing_returns_none(self, opts):
        assert read_batch_state("does-not-exist", opts) is None

    def test_read_corrupt_returns_none(self, opts, tmp_path):
        # Write garbage to the expected location.
        write_batch_state("JID", {"ok": True}, opts)
        # Overwrite with non-msgpack bytes.
        import salt.utils.files
        from salt.utils.batch_state import _batch_state_path

        path = _batch_state_path("JID", opts)
        with salt.utils.files.fopen(path, "wb") as fp:
            fp.write(b"\xff\xff\xff\xff not msgpack \xff\xff")
        assert read_batch_state("JID", opts) is None


class TestActiveIndexPersistence:
    def test_round_trip_set(self, opts):
        write_active_index({"JID1", "JID2"}, opts)
        assert read_active_index(opts) == {"JID1", "JID2"}

    def test_read_missing_returns_empty_set(self, opts):
        assert read_active_index(opts) == set()

    def test_add_and_remove(self, opts):
        add_to_active_index("JID1", opts)
        add_to_active_index("JID2", opts)
        assert read_active_index(opts) == {"JID1", "JID2"}

        remove_from_active_index("JID1", opts)
        assert read_active_index(opts) == {"JID2"}

    def test_remove_missing_is_noop(self, opts):
        add_to_active_index("JID1", opts)
        remove_from_active_index("JID-does-not-exist", opts)
        assert read_active_index(opts) == {"JID1"}



"""
Unit tests for worker pools configuration
"""

import pytest

from salt.config.worker_pools import (
    DEFAULT_WORKER_POOLS,
    get_worker_pools_config,
    validate_worker_pools_config,
)


class TestWorkerPoolsConfig:
    """Test worker pools configuration functions"""

    def test_default_worker_pools_structure(self):
        """Test that DEFAULT_WORKER_POOLS has correct structure"""
        assert isinstance(DEFAULT_WORKER_POOLS, dict)
        assert "default" in DEFAULT_WORKER_POOLS
        assert DEFAULT_WORKER_POOLS["default"]["worker_count"] == 5
        assert DEFAULT_WORKER_POOLS["default"]["commands"] == ["*"]

    def test_get_worker_pools_config_default(self):
        """Test get_worker_pools_config with default config"""
        opts = {"worker_pools_enabled": True, "worker_pools": {}}
        result = get_worker_pools_config(opts)
        assert result == DEFAULT_WORKER_POOLS

    def test_get_worker_pools_config_disabled(self):
        """Test get_worker_pools_config when pools are disabled"""
        opts = {"worker_pools_enabled": False}
        result = get_worker_pools_config(opts)
        assert result is None

    def test_get_worker_pools_config_worker_threads_compat(self):
        """Test backward compatibility with worker_threads"""
        opts = {"worker_pools_enabled": True, "worker_threads": 10, "worker_pools": {}}
        result = get_worker_pools_config(opts)
        assert result == {"default": {"worker_count": 10, "commands": ["*"]}}

    def test_get_worker_pools_config_custom(self):
        """Test get_worker_pools_config with custom pools"""
        custom_pools = {
            "fast": {"worker_count": 2, "commands": ["ping"]},
            "slow": {"worker_count": 3, "commands": ["*"]},
        }
        opts = {"worker_pools_enabled": True, "worker_pools": custom_pools}
        result = get_worker_pools_config(opts)
        assert result == custom_pools

    def test_validate_worker_pools_config_valid_default(self):
        """Test validation with valid default config"""
        opts = {"worker_pools_enabled": True, "worker_pools": DEFAULT_WORKER_POOLS}
        assert validate_worker_pools_config(opts) is True

    def test_validate_worker_pools_config_valid_catchall(self):
        """Test validation with valid catchall pool"""
        opts = {
            "worker_pools_enabled": True,
            "worker_pools": {
                "fast": {"worker_count": 2, "commands": ["ping"]},
                "slow": {"worker_count": 3, "commands": ["*"]},
            },
        }
        assert validate_worker_pools_config(opts) is True

    def test_validate_worker_pools_config_duplicate_catchall(self):
        """Test validation catches duplicate catchall"""
        opts = {
            "worker_pools_enabled": True,
            "worker_pools": {
                "pool1": {"worker_count": 2, "commands": ["*"]},
                "pool2": {"worker_count": 3, "commands": ["*"]},
            },
        }
        with pytest.raises(ValueError, match="Multiple pools have catchall"):
            validate_worker_pools_config(opts)

    def test_validate_worker_pools_config_duplicate_command(self):
        """Test validation catches duplicate commands"""
        opts = {
            "worker_pools_enabled": True,
            "worker_pools": {
                "pool1": {"worker_count": 2, "commands": ["ping", "*"]},
                "pool2": {"worker_count": 3, "commands": ["ping"]},
            },
        }
        with pytest.raises(ValueError, match="Command 'ping' mapped to multiple pools"):
            validate_worker_pools_config(opts)

    def test_validate_worker_pools_config_invalid_worker_count(self):
        """Test validation catches invalid worker_count"""
        opts = {
            "worker_pools_enabled": True,
            "worker_pools": {
                "pool1": {"worker_count": 0, "commands": ["*"]},
            },
        }
        with pytest.raises(ValueError, match="worker_count must be integer >= 1"):
            validate_worker_pools_config(opts)

    def test_validate_worker_pools_config_no_catchall(self):
        """Test validation requires a catchall pool"""
        opts = {
            "worker_pools_enabled": True,
            "worker_pools": {
                "pool1": {"worker_count": 2, "commands": ["ping"]},
            },
        }
        with pytest.raises(ValueError, match="No catchall pool"):
            validate_worker_pools_config(opts)

    def test_validate_worker_pools_config_disabled(self):
        """Test validation passes when pools are disabled"""
        opts = {"worker_pools_enabled": False}
        assert validate_worker_pools_config(opts) is True



import multiprocessing
import subprocess

import salt.utils.platform
from tests.support.mock import patch


def test_linux_distribution():
    """
    Test that when `distro` fails with a `subprocess.CalledProcessError` salt
    returns empty strings as default values.
    """
    distro_name = "Salt"
    distro_version = "1"
    distro_codename = "Awesome"
    with patch("distro.name", return_value=distro_name):
        with patch("distro.version", return_value=distro_version), patch(
            "distro.codename", return_value=distro_codename
        ):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )

        distro_version = ""
        with patch(
            "distro.version",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ), patch("distro.codename", return_value=distro_codename):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )
        distro_codename = ""
        with patch(
            "distro.version",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ), patch(
            "distro.codename",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )


def test_spawning_platform_spawn():
    """
    spawning_platform() must return True when the multiprocessing start method
    is "spawn" (Windows default, macOS default on Python >= 3.8).
    """
    with patch.object(multiprocessing, "get_start_method", return_value="spawn"):
        assert salt.utils.platform.spawning_platform() is True


def test_spawning_platform_forkserver():
    """
    spawning_platform() must return True when the multiprocessing start method
    is "forkserver".  Like "spawn", forkserver transfers the Process object to
    the child via pickle, so Salt must prepare __getstate__/__setstate__ for it.
    This is the Linux default starting with Python 3.14.
    """
    with patch.object(multiprocessing, "get_start_method", return_value="forkserver"):
        assert salt.utils.platform.spawning_platform() is True


def test_spawning_platform_fork():
    """
    spawning_platform() must return False when the multiprocessing start method
    is "fork" (Linux default on Python < 3.14).  Fork inherits process state
    directly, so pickling is not required.
    """
    with patch.object(multiprocessing, "get_start_method", return_value="fork"):
        assert salt.utils.platform.spawning_platform() is False



import asyncio
import contextlib
import copy
import logging
import os
import pathlib
import signal
import time
import uuid

import pytest
import tornado
import tornado.gen
import tornado.testing

import salt.minion
import salt.modules.test as test_mod
import salt.syspaths
import salt.utils.crypt
import salt.utils.jid
import salt.utils.platform
import salt.utils.process
from salt._compat import ipaddress
from salt.exceptions import SaltClientError, SaltMasterUnresolvableError, SaltSystemExit
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def connect_master_mock():
    class ConnectMasterMock:
        """
        Mock connect master call.

        The first call will raise an exception stored on the exc attribute.
        Subsequent calls will return True.
        """

        def __init__(self):
            self.calls = 0
            self.exc = Exception

        @tornado.gen.coroutine
        def __call__(self, *args, **kwargs):
            self.calls += 1
            if self.calls == 1:
                raise self.exc()
            else:
                return True

    return ConnectMasterMock()


def test_minion_load_grains_false(minion_opts):
    """
    Minion does not generate grains when load_grains is False
    """
    minion_opts["grains"] = {"foo": "bar"}
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts, load_grains=False)
        try:
            assert minion.opts["grains"] == minion_opts["grains"]
            grainsfunc.assert_not_called()
        finally:
            minion.destroy()


def test_minion_load_grains_true(minion_opts):
    """
    Minion generates grains when load_grains is True
    """
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts, load_grains=True)
        try:
            assert minion.opts["grains"] != {}
            grainsfunc.assert_called()
        finally:
            minion.destroy()


def test_minion_load_grains_default(minion_opts):
    """
    Minion load_grains defaults to True
    """
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts)
        try:
            assert minion.opts["grains"] != {}
            grainsfunc.assert_called()
        finally:
            minion.destroy()


@pytest.mark.parametrize(
    "event",
    [
        (
            "fire_event",
            lambda data, tag, cb=None, timeout=60: True,
        ),
        (
            "fire_event_async",
            lambda data, tag, cb=None, timeout=60: tornado.gen.maybe_future(True),
        ),
    ],
)
def test_send_req_fires_completion_event(event, minion_opts):
    req_id = uuid.uuid4()
    event_enter = MagicMock()
    event_enter.send.side_effect = event[1]
    event = MagicMock()
    event.__enter__.return_value = event_enter

    with patch("salt.utils.event.get_event", return_value=event), patch(
        "uuid.uuid4", return_value=req_id
    ):
        minion_opts["random_startup_delay"] = 0
        minion_opts["return_retry_tries"] = 30
        minion_opts["grains"] = {}
        with patch("salt.loader.grains"):
            minion = salt.minion.Minion(minion_opts)

            try:
                load = {"load": "value"}
                timeout = 60

                # XXX This is buggy because "async" in event[0] will never evaluate
                # to True and if it *did* evaluate to true the test would fail
                # because you Mock isn't a co-routine.
                if "async" in event[0]:
                    rtn = minion._send_req_async(load, timeout).result()
                else:
                    rtn = minion._send_req_sync(load, timeout)

                fire_event_called = False
                # get the
                for idx, call in enumerate(event.mock_calls, 1):
                    if "fire_event" in call[0]:
                        condition_event_tag = (
                            len(call.args) > 1
                            and call.args[1]
                            == f"__master_req_channel_payload/{req_id}/{minion_opts['master']}"
                        )
                        condition_event_tag_error = (
                            "{} != {}; Call(number={}): {}".format(
                                idx, call, call.args[1], "__master_req_channel_payload"
                            )
                        )
                        condition_timeout = (
                            len(call.kwargs) == 1 and call.kwargs["timeout"] == timeout
                        )
                        condition_timeout_error = (
                            "{} != {}; Call(number={}): {}".format(
                                idx, call, call.kwargs["timeout"], timeout
                            )
                        )

                        fire_event_called = True
                        assert condition_event_tag, condition_event_tag_error
                        assert condition_timeout, condition_timeout_error

                assert fire_event_called
                assert rtn
            finally:
                minion.destroy()


async def test_send_req_async_regression_62453(minion_opts):

    class MockEvent:

        def __init__(self, *args, **kwargs):
            pass

        @tornado.gen.coroutine
        def fire_event_async(self, *args, **kwargs):
            return

        def get_event(self, *args, **kwargs):
            return

        def __enter__(self):
            return self

        def __exit__(self, *args):
            return

    def get_event(*args, **kwargs):
        return MockEvent()

    minion_opts["random_startup_delay"] = 0
    minion_opts["return_retry_tries"] = 5
    minion_opts["grains"] = {}
    minion_opts["ipc_mode"] = "tcp"
    with patch("salt.loader.grains"):
        minion = salt.minion.Minion(minion_opts)

        load = {"load": "value"}
        timeout = 1

        with patch("salt.utils.event.get_event", get_event):
            # We are just validating no exception is raised
            with pytest.raises(TimeoutError):
                rtn = await minion._send_req_async(load, timeout)


def test_mine_send_tries(minion_opts):
    channel_enter = MagicMock()
    channel_enter.send.side_effect = lambda load, timeout, tries: tries
    channel = MagicMock()
    channel.__enter__.return_value = channel_enter

    minion_opts["return_retry_tries"] = 20
    with patch("salt.channel.client.ReqChannel.factory", return_value=channel), patch(
        "salt.loader.grains"
    ):
        minion = salt.minion.Minion(minion_opts)
        minion.tok = "token"

        data = {}
        tag = "tag"

        rtn = minion._mine_send(tag, data)
        assert rtn == 20


def test_invalid_master_address(minion_opts):
    minion_opts.update(
        {
            "ipv6": False,
            "master": float("127.0"),
            "master_port": "4555",
            "retry_dns": False,
        }
    )
    with pytest.raises(SaltSystemExit):
        salt.minion.resolve_dns(minion_opts)


def test_source_int_name_local(minion_opts):
    """
    test when file_client local and
    source_interface_name is set
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": True,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "local",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "master_ip": "127.0.0.1",
            "source_ip": "111.1.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
        }


@pytest.mark.slow_test
def test_source_int_name_remote(minion_opts):
    """
    test when file_client remote and
    source_interface_name is set and
    interface is down
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "remote",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "master_ip": "127.0.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
        }


@pytest.mark.slow_test
def test_source_address(minion_opts):
    """
    test when source_address is set
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "local",
            "source_interface_name": "",
            "source_address": "111.1.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "source_publish_port": 49018,
            "source_ret_port": 49017,
            "master_uri": "tcp://127.0.0.1:4555",
            "source_ip": "111.1.0.1",
            "master_ip": "127.0.0.1",
        }


# Tests for _handle_decoded_payload in the salt.minion.Minion() class: 3
@pytest.mark.slow_test
async def test_handle_decoded_payload_jid_match_in_jid_queue(minion_opts, io_loop):
    """
    Tests that the _handle_decoded_payload function returns when a jid is given that is already present
    in the jid_queue.

    Note: This test doesn't contain all of the patch decorators above the function like the other tests
    for _handle_decoded_payload below. This is essential to this test as the call to the function must
    return None BEFORE any of the processes are spun up because we should be avoiding firing duplicate
    jobs.
    """
    mock_data = {"fun": "foo.bar", "jid": 123}
    mock_jid_queue = [123]
    minion = salt.minion.Minion(
        minion_opts,
        jid_queue=copy.copy(mock_jid_queue),
        io_loop=io_loop,
    )
    try:
        ret = await minion._handle_decoded_payload(mock_data)
        assert minion.jid_queue == mock_jid_queue
        assert ret is None
    finally:
        minion.destroy()


@pytest.mark.slow_test
async def test_handle_decoded_payload_jid_queue_addition(minion_opts, io_loop):
    """
    Tests that the _handle_decoded_payload function adds a jid to the minion's jid_queue when the new
    jid isn't already present in the jid_queue.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        mock_jid = 11111
        mock_data = {"fun": "foo.bar", "jid": mock_jid}
        mock_jid_queue = [123, 456]
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=io_loop,
        )
        try:

            # Assert that the minion's jid_queue attribute matches the mock_jid_queue as a baseline
            # This can help debug any test failures if the _handle_decoded_payload call fails.
            assert minion.jid_queue == mock_jid_queue

            # Call the _handle_decoded_payload function and update the mock_jid_queue to include the new
            # mock_jid. The mock_jid should have been added to the jid_queue since the mock_jid wasn't
            # previously included. The minion's jid_queue attribute and the mock_jid_queue should be equal.
            await minion._handle_decoded_payload(mock_data)
            mock_jid_queue.append(mock_jid)
            assert minion.jid_queue == mock_jid_queue
        finally:
            minion.destroy()


@pytest.mark.slow_test
async def test_handle_decoded_payload_jid_queue_reduced_minion_jid_queue_hwm(
    minion_opts, io_loop
):
    """
    Tests that the _handle_decoded_payload function removes a jid from the minion's jid_queue when the
    minion's jid_queue high water mark (minion_jid_queue_hwm) is hit.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["minion_jid_queue_hwm"] = 2
        mock_data = {"fun": "foo.bar", "jid": 789}
        mock_jid_queue = [123, 456]
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=io_loop,
        )
        try:

            # Assert that the minion's jid_queue attribute matches the mock_jid_queue as a baseline
            # This can help debug any test failures if the _handle_decoded_payload call fails.
            assert minion.jid_queue == mock_jid_queue

            # Call the _handle_decoded_payload function and check that the queue is smaller by one item
            # and contains the new jid
            await minion._handle_decoded_payload(mock_data)
            assert len(minion.jid_queue) == 2
            assert minion.jid_queue == [456, 789]
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_process_count_max(minion_opts, io_loop):
    """
    Tests that the _handle_decoded_payload function does not spawn more than the configured amount of processes,
    as per process_count_max.
    """
    start_mock = MagicMock(return_value=True)

    def mock_proc_side_effect(*args, **kwargs):
        m = MagicMock(name="MockProcess")
        m.is_alive.return_value = True
        m.start = start_mock
        return m

    @contextlib.asynccontextmanager
    async def mock_await_lock(*args, **kwargs):
        yield

    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.SignalHandlingProcess",
        MagicMock(side_effect=mock_proc_side_effect),
    ), patch(
        "salt.minion.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), patch(
        "os.makedirs", MagicMock()
    ), patch(
        "salt.utils.files.fopen", MagicMock()
    ), patch(
        "salt.payload.dump", MagicMock()
    ), patch(
        "salt.utils.files.await_lock", side_effect=mock_await_lock
    ), patch(
        "salt.loader.grains", MagicMock(return_value={"id": "foo", "os": "Linux"})
    ):
        process_count_max = 10
        minion_opts["__role"] = "minion"
        minion_opts["minion_jid_queue_hwm"] = 100
        minion_opts["process_count_max"] = process_count_max
        # cachedir needed for lock
        minion_opts["cachedir"] = "/tmp/salt_test_cache"

        minion = salt.minion.Minion(minion_opts, jid_queue=[], io_loop=io_loop)
        try:
            # up until process_count_max: processes are started normally
            for i in range(process_count_max):
                mock_data = {"fun": "foo.bar", "jid": str(i)}
                io_loop.run_sync(
                    lambda data=mock_data: minion._handle_decoded_payload(data)
                )
                assert start_mock.call_count == i + 1
                assert len(minion.jid_queue) == i + 1

            # above process_count_max: Queue logic kicks in
            mock_data = {"fun": "foo.bar", "jid": str(process_count_max + 1)}

            # Run execution
            io_loop.run_sync(lambda: minion._handle_decoded_payload(mock_data))

            # Assert NO new process started
            assert start_mock.call_count == process_count_max
            # Assert Job was queued (payload dumped)
            assert salt.payload.dump.called
            # Assert JID added to active queue (deduplication cache)
            assert len(minion.jid_queue) == process_count_max + 1

        finally:
            minion.destroy()


async def test_process_queue_rechecks_count_per_job(minion_opts):
    """
    Test that job queue processing re-checks process count before each individual job,
    preventing race conditions where process count changes during batch processing.
    """
    # Create a simple test that just verifies the queue processing method exists
    from salt.minion import Minion

    minion = Minion(minion_opts)
    try:
        # Just test that the method exists and can be called without crashing
        await minion._process_process_queue_async_impl()
        # If we get here without exception, test passes
        assert True
    finally:
        minion.destroy()


def test_cleanup_orphaned_queue_files(minion_opts):
    """
    Test that orphaned running_ queue files are cleaned up on minion startup.
    This prevents stale files from blocking future jobs after minion crashes.
    """
    # Create a simple test that just verifies the method exists and can be called
    from salt.minion import Minion

    minion = Minion(minion_opts)
    try:
        # Just test that the method exists and doesn't crash when called
        minion._cleanup_orphaned_queue_files()
        # If we get here without exception, test passes
        assert True
    finally:
        minion.destroy()


@pytest.mark.slow_test
async def test_beacons_before_connect(minion_opts):
    """
    Tests that the 'beacons_before_connect' option causes the beacons to be initialized before connect.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["beacons_before_connect"] = True
        io_loop = tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:

            try:
                await minion.tune_in(start=True)
            except RuntimeError:
                pass

            # Make sure beacons are initialized but the sheduler is not
            assert "beacons" in minion.periodic_callbacks
            assert "schedule" not in minion.periodic_callbacks
        finally:
            minion.destroy()


@pytest.mark.slow_test
async def test_scheduler_before_connect(minion_opts):
    """
    Tests that the 'scheduler_before_connect' option causes the scheduler to be initialized before connect.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["scheduler_before_connect"] = True
        io_loop = tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:
            try:
                await minion.tune_in(start=True)
            except RuntimeError:
                pass

            # Make sure the scheduler is initialized but the beacons are not
            assert "schedule" in minion.periodic_callbacks
            assert "beacons" not in minion.periodic_callbacks
        finally:
            minion.destroy()


def test_minion_module_refresh(minion_opts):
    """
    Tests that the 'module_refresh' just return in case there is no 'schedule'
    because destroy method was already called.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=tornado.ioloop.IOLoop(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )
            assert hasattr(minion, "schedule")
            minion.destroy()
            assert not hasattr(minion, "schedule")
            assert not minion.module_refresh()
        finally:
            minion.destroy()


def test_minion_module_refresh_beacons_refresh(minion_opts):
    """
    Tests that 'module_refresh' calls beacons_refresh and that the
    minion object has a beacons attribute with beacons.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=tornado.ioloop.IOLoop(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )
            assert not hasattr(minion, "beacons")
            minion.module_refresh()
            assert hasattr(minion, "beacons")
            assert hasattr(minion.beacons, "beacons")
            assert "service.beacon" in minion.beacons.beacons
            minion.destroy()
        finally:
            if minion is not None:
                minion.destroy()


def test_beacons_refresh_preserves_interval_map(minion_opts):
    """
    Tests that 'beacons_refresh' preserves the interval_map so that
    beacon intervals are not reset during module refresh.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion = None
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=tornado.ioloop.IOLoop.current(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )

            minion.module_refresh()
            assert hasattr(minion, "beacons")
            assert hasattr(minion.beacons, "interval_map")

            test_interval_map = {"status": 50, "diskusage": 30}
            minion.beacons.interval_map = test_interval_map.copy()

            old_beacons = minion.beacons

            minion.beacons_refresh()

            assert minion.beacons is not old_beacons

            assert minion.beacons.interval_map == test_interval_map
            assert minion.beacons.interval_map["status"] == 50
            assert minion.beacons.interval_map["diskusage"] == 30

        finally:
            if minion is not None:
                minion.destroy()


def test_beacons_refresh_closes_old_beacons(minion_opts):
    """
    Tests that 'beacons_refresh' calls close_beacons() on the old Beacon
    instance before replacing it, preventing inotify fd leaks.

    See: https://github.com/saltstack/salt/issues/66449
    See: https://github.com/saltstack/salt/issues/58907
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion = None
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=tornado.ioloop.IOLoop.current(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )

            minion.module_refresh()
            assert hasattr(minion, "beacons")

            old_beacons = minion.beacons
            with patch.object(old_beacons, "close_beacons") as close_mock:
                minion.beacons_refresh()
                close_mock.assert_called_once()

            assert minion.beacons is not old_beacons

        finally:
            if minion is not None:
                minion.destroy()


@pytest.mark.slow_test
async def test_when_ping_interval_is_set_the_callback_should_be_added_to_periodic_callbacks(
    minion_opts,
):
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["ping_interval"] = 10
        io_loop = tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:
            try:
                minion.connected = MagicMock(side_effect=(False, True))

                # _fire_master_minion_start is now called as a coroutine via create_task
                # so it must be an async function
                async def async_mock():
                    pass

                minion._fire_master_minion_start = async_mock
                minion.tune_in(start=False)
            except RuntimeError:
                pass

            # Make sure the scheduler is initialized but the beacons are not
            assert "ping" in minion.periodic_callbacks
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_when_passed_start_event_grains(minion_opts):
    # provide mock opts an os grain since we'll look for it later.
    minion_opts["grains"]["os"] = "linux"
    minion_opts["start_event_grains"] = ["os"]
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master(
            "Minion has started", "minion_start", include_startup_grains=True
        )
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" in load
        assert "os" in load["grains"]
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_when_not_passed_start_event_grains(minion_opts):
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master("Minion has started", "minion_start")
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_when_other_events_fired_and_start_event_grains_are_set(minion_opts):
    minion_opts["start_event_grains"] = ["os"]
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master("Custm_event_fired", "custom_event")
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_minimal_payload(minion_opts):
    """
    A minimal published job with start_event=True should produce a
    salt/job/<jid>/start/<minion_id> event whose payload contains the
    core identifying fields and excludes the function arguments.
    """
    minion_opts["id"] = "minion-under-test"
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        data = {
            "jid": "20260429000000000000",
            "fun": "test.ping",
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
            "arg": ["should-not-leak"],
        }
        minion._fire_start_event(data)
        minion._fire_master.assert_called_once()
        load, tag = minion._fire_master.call_args[0]
        assert tag == "salt/job/20260429000000000000/start/{}".format(minion_opts["id"])
        assert load["id"] == minion_opts["id"]
        assert load["jid"] == data["jid"]
        assert load["fun"] == data["fun"]
        assert load["tgt"] == data["tgt"]
        assert load["tgt_type"] == data["tgt_type"]
        assert load["user"] == data["user"]
        assert "arg" not in load
        assert "fun_args" not in load
        assert "master_id" not in load
        assert "metadata" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_includes_master_id_and_metadata(minion_opts):
    """
    When the published load carries master_id and metadata, the start
    event should propagate both.
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        data = {
            "jid": "20260429000000000001",
            "fun": "test.ping",
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
            "master_id": "master-a",
            "metadata": {"ticket": "INC-1234"},
        }
        minion._fire_start_event(data)
        load, _tag = minion._fire_master.call_args[0]
        assert load["master_id"] == "master-a"
        assert load["metadata"] == {"ticket": "INC-1234"}
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_swallows_failures(minion_opts):
    """
    A failure inside _fire_master must not propagate out of
    _fire_start_event; firing the start event must never abort job
    execution.
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock(side_effect=RuntimeError("transport down"))
        data = {
            "jid": "20260429000000000002",
            "fun": "test.ping",
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
        }
        minion._fire_start_event(data)
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_empty_metadata_is_propagated(minion_opts):
    """
    An empty metadata dict is still meaningful (caller deliberately sent
    one) and must be propagated. Distinguishes ``metadata={}`` from a
    missing key.
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        data = {
            "jid": "20260429000000000010",
            "fun": "test.ping",
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
            "metadata": {},
        }
        minion._fire_start_event(data)
        load, _tag = minion._fire_master.call_args[0]
        assert "metadata" in load
        assert load["metadata"] == {}
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_omits_falsy_master_id(minion_opts):
    """
    A master_id of None or empty string must not appear in the start
    event load (the gate is truthiness, matching how master_id flows in
    the publish path).
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        for falsy in (None, ""):
            minion._fire_master.reset_mock()
            data = {
                "jid": "20260429000000000011",
                "fun": "test.ping",
                "tgt": "*",
                "tgt_type": "glob",
                "user": "root",
                "master_id": falsy,
            }
            minion._fire_start_event(data)
            load, _tag = minion._fire_master.call_args[0]
            assert "master_id" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_fire_start_event_multi_fun_passes_list(minion_opts):
    """
    For multi-fun jobs, ``data["fun"]`` is a list. The start event
    should still fire successfully and propagate the list of function
    names verbatim.
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        data = {
            "jid": "20260429000000000012",
            "fun": ["test.ping", "test.echo"],
            "arg": [[], ["hello"]],
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
        }
        minion._fire_start_event(data)
        minion._fire_master.assert_called_once()
        load, _tag = minion._fire_master.call_args[0]
        assert load["fun"] == ["test.ping", "test.echo"]
        assert "arg" not in load
    finally:
        minion.destroy()


def _make_thread_return_minion_mock(tmp_path, opts):
    """
    Build a MagicMock that quacks like a Minion well enough for
    _thread_return to reach (or skip) the start-event gate. Heavy
    dependencies (executors, returners, the actual function) are
    short-circuited; only the start-event gating logic is exercised.
    """
    proc_dir = tmp_path / "proc"
    proc_dir.mkdir(exist_ok=True)
    minion_instance = MagicMock()
    minion_instance.proc_dir = str(proc_dir)
    minion_instance.opts = opts
    minion_instance.executors = {}
    minion_instance.module_executors = []
    minion_instance.functions = MagicMock()
    minion_instance.functions.__contains__.return_value = True
    minion_instance.functions.pack = {"__context__": {"retcode": 0}}
    minion_instance._execute_job_function = MagicMock(return_value="ok")
    minion_instance._return_pub = MagicMock()
    minion_instance._fire_master = MagicMock()
    minion_instance._fire_start_event = MagicMock()
    minion_instance._return_retry_timer = MagicMock(return_value=1)
    minion_instance.connected = False  # skip _return_pub
    minion_instance.returners = MagicMock()
    minion_instance.function_errors = {}
    return minion_instance


@pytest.mark.slow_test
def test_thread_return_fires_start_event_when_requested(tmp_path, minion_opts):
    """
    _thread_return must invoke _fire_start_event exactly once when the
    published load carries start_event=True.
    """
    minion_opts["multiprocessing"] = False
    minion_opts["id"] = "minion-thread-return"
    minion_instance = _make_thread_return_minion_mock(tmp_path, minion_opts)
    data = {
        "jid": "20260429000000000020",
        "fun": "test.ping",
        "arg": [],
        "tgt": "*",
        "tgt_type": "glob",
        "user": "root",
        "ret": "",
        "start_event": True,
    }
    salt.minion.Minion._thread_return(minion_instance, minion_opts, data)
    minion_instance._fire_start_event.assert_called_once_with(data)


@pytest.mark.slow_test
def test_thread_return_skips_start_event_when_not_requested(tmp_path, minion_opts):
    """
    _thread_return must NOT invoke _fire_start_event when the published
    load has no start_event key. This is the default behavior for all
    pre-existing callers and must not regress.
    """
    minion_opts["multiprocessing"] = False
    minion_opts["id"] = "minion-thread-return"
    minion_instance = _make_thread_return_minion_mock(tmp_path, minion_opts)
    data = {
        "jid": "20260429000000000021",
        "fun": "test.ping",
        "arg": [],
        "tgt": "*",
        "tgt_type": "glob",
        "user": "root",
        "ret": "",
    }
    salt.minion.Minion._thread_return(minion_instance, minion_opts, data)
    minion_instance._fire_start_event.assert_not_called()


@pytest.mark.slow_test
def test_thread_return_skips_start_event_when_falsy(tmp_path, minion_opts):
    """
    A start_event of False/None/empty string must be treated as
    opt-out; _fire_start_event should not be invoked.
    """
    minion_opts["multiprocessing"] = False
    minion_opts["id"] = "minion-thread-return"
    for falsy in (False, None, "", 0):
        minion_instance = _make_thread_return_minion_mock(tmp_path, minion_opts)
        data = {
            "jid": "20260429000000000022",
            "fun": "test.ping",
            "arg": [],
            "tgt": "*",
            "tgt_type": "glob",
            "user": "root",
            "ret": "",
            "start_event": falsy,
        }
        salt.minion.Minion._thread_return(minion_instance, minion_opts, data)
        assert (
            minion_instance._fire_start_event.call_count == 0
        ), f"Unexpectedly fired start event for falsy value {falsy!r}"


def _make_thread_multi_return_minion_mock(tmp_path, opts):
    """
    Sibling helper for _thread_multi_return tests.
    """
    proc_dir = tmp_path / "proc"
    proc_dir.mkdir(exist_ok=True)
    minion_instance = MagicMock()
    minion_instance.proc_dir = str(proc_dir)
    minion_instance.opts = opts
    minion_instance.executors = {}
    minion_instance.module_executors = []
    minion_instance.functions = MagicMock()
    minion_instance.functions.__contains__.return_value = True
    minion_instance.functions.pack = {"__context__": {"retcode": 0}}
    minion_instance._execute_job_function = MagicMock(return_value="ok")
    minion_instance._return_pub_multi = MagicMock()
    minion_instance._fire_master = MagicMock()
    minion_instance._fire_start_event = MagicMock()
    minion_instance._return_retry_timer = MagicMock(return_value=1)
    minion_instance.connected = False
    minion_instance.returners = MagicMock()
    minion_instance.function_errors = {}
    return minion_instance


@pytest.mark.slow_test
def test_thread_multi_return_fires_start_event_once_per_jid(tmp_path, minion_opts):
    """
    Multi-fun jobs share a single jid; the start event must fire
    exactly once for the jid regardless of how many sub-functions are
    in the load.
    """
    minion_opts["multiprocessing"] = False
    minion_opts["id"] = "minion-multi"
    minion_instance = _make_thread_multi_return_minion_mock(tmp_path, minion_opts)
    data = {
        "jid": "20260429000000000030",
        "fun": ["test.ping", "test.echo", "test.version"],
        "arg": [[], ["hello"], []],
        "tgt": "*",
        "tgt_type": "glob",
        "user": "root",
        "ret": "",
        "start_event": True,
    }
    salt.minion.Minion._thread_multi_return(minion_instance, minion_opts, data)
    minion_instance._fire_start_event.assert_called_once_with(data)


@pytest.mark.slow_test
def test_thread_multi_return_skips_start_event_when_not_requested(
    tmp_path, minion_opts
):
    """
    Multi-fun jobs without start_event must not produce a start event,
    matching the single-fun gating.
    """
    minion_opts["multiprocessing"] = False
    minion_opts["id"] = "minion-multi"
    minion_instance = _make_thread_multi_return_minion_mock(tmp_path, minion_opts)
    data = {
        "jid": "20260429000000000031",
        "fun": ["test.ping", "test.echo"],
        "arg": [[], []],
        "tgt": "*",
        "tgt_type": "glob",
        "user": "root",
        "ret": "",
    }
    salt.minion.Minion._thread_multi_return(minion_instance, minion_opts, data)
    minion_instance._fire_start_event.assert_not_called()


@pytest.mark.slow_test
def test_fire_start_event_missing_jid_does_not_raise(minion_opts):
    """
    A malformed payload (missing jid) must not propagate an exception
    out of _fire_start_event. Job execution must never be aborted by a
    failure in the start-event helper.
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._fire_master = MagicMock()
        data = {"fun": "test.ping"}
        # Should not raise even though "jid" is required and missing.
        minion._fire_start_event(data)
        # _fire_master should not have been called because payload
        # construction failed before reaching it.
        minion._fire_master.assert_not_called()
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_minion_retry_dns_count(minion_opts):
    """
    Tests that the resolve_dns will retry dns look ups for a maximum of
    3 times before raising a SaltMasterUnresolvableError exception.
    """
    minion_opts.update(
        {
            "ipv6": False,
            "master": "dummy",
            "master_port": "4555",
            "retry_dns": 1,
            "retry_dns_count": 3,
        },
    )
    with pytest.raises(SaltMasterUnresolvableError):
        salt.minion.resolve_dns(minion_opts)


@pytest.mark.slow_test
def test_gen_modules_executors(minion_opts):
    """
    Ensure gen_modules is called with the correct arguments #54429
    """
    io_loop = tornado.ioloop.IOLoop()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)

    class MockPillarCompiler:
        def compile_pillar(self):
            return {}

    try:
        with patch("salt.pillar.get_pillar", return_value=MockPillarCompiler()):
            with patch("salt.loader.executors", mock=MagicMock()) as execmock:
                minion.gen_modules()
        execmock.assert_called_once_with(
            minion.opts, functions=minion.functions, proxy=minion.proxy, context={}
        )
    finally:
        minion.destroy()


def test_minion_manage_schedule(minion_opts):
    """
    Tests that the manage_schedule will call the add function, adding
    schedule data into opts.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        io_loop = tornado.ioloop.IOLoop()

        with patch("salt.utils.schedule.clean_proc_dir", MagicMock(return_value=None)):
            try:
                mock_functions = {"test.ping": None}

                minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
                minion.schedule = salt.utils.schedule.Schedule(
                    minion_opts,
                    mock_functions,
                    returners={},
                    new_instance=True,
                )

                minion.opts["foo"] = "bar"
                schedule_data = {
                    "test_job": {
                        "function": "test.ping",
                        "return_job": False,
                        "jid_include": True,
                        "maxrunning": 2,
                        "seconds": 10,
                    }
                }

                data = {
                    "name": "test-item",
                    "schedule": schedule_data,
                    "func": "add",
                    "persist": False,
                }
                tag = "manage_schedule"

                minion.manage_schedule(tag, data)
                assert "test_job" in minion.opts["schedule"]
            finally:
                del minion.schedule
                minion.destroy()
                del minion


def test_minion_manage_beacons(minion_opts):
    """
    Tests that the manage_beacons will call the add function, adding
    beacon data into opts.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion = None
        try:
            minion_opts["beacons"] = {}

            # io_loop must be a real Tornado IOLoop because our code calls
            # salt.utils.asynchronous.aioloop() on it
            io_loop = tornado.ioloop.IOLoop()

            mock_functions = {"test.ping": None}
            minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
            minion.beacons = salt.beacons.Beacon(minion_opts, mock_functions)

            bdata = [{"salt-master": "stopped"}, {"apache2": "stopped"}]
            data = {"name": "ps", "beacon_data": bdata, "func": "add"}

            tag = "manage_beacons"
            log.debug("==== minion.opts %s ====", minion.opts)

            minion.manage_beacons(tag, data)
            assert "ps" in minion.opts["beacons"]
            assert minion.opts["beacons"]["ps"] == bdata
        finally:
            if minion is not None:
                minion.destroy()


def test_prep_ip_port():
    _ip = ipaddress.ip_address

    opts = {"master": "10.10.0.3", "master_uri_format": "ip_only"}
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": _ip("10.10.0.3")}

    opts = {
        "master": "10.10.0.3",
        "master_port": 1234,
        "master_uri_format": "default",
    }
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": "10.10.0.3"}

    opts = {"master": "10.10.0.3:1234", "master_uri_format": "default"}
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": "10.10.0.3", "master_port": 1234}

    opts = {"master": "host name", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)

    opts = {"master": "10.10.0.3:abcd", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)

    opts = {"master": "10.10.0.3::1234", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)


@pytest.mark.skip_on_windows(reason="Skippin, no Salt master running on Windows.")
async def test_master_type_failover(minion_opts):
    """
    Tests master_type "failover" to not fall back to 127.0.0.1 address when master does not resolve in DNS
    """
    minion_opts.update(
        {
            "master_type": "failover",
            "master": ["master1", "master2"],
            "__role": "",
            "retry_dns": 0,
        }
    )

    class MockPubChannel:
        def connect(self):
            raise SaltClientError("MockedChannel")

        def close(self):
            return

    def mock_resolve_dns(opts, fallback=False):
        assert not fallback

        if opts["master"] == "master1":
            raise SaltClientError("Cannot resolve {}".format(opts["master"]))

        return {
            "master_ip": "192.168.2.1",
            "master_uri": "tcp://192.168.2.1:4505",
        }

    def mock_channel_factory(opts, **kwargs):
        assert opts["master"] == "master2"
        return MockPubChannel()

    with patch("salt.minion.resolve_dns", mock_resolve_dns), patch(
        "salt.channel.client.AsyncPubChannel.factory", mock_channel_factory
    ), patch("salt.loader.grains", MagicMock(return_value=[])):
        with pytest.raises(SaltClientError):
            minion = salt.minion.Minion(minion_opts)
            await minion.connect_master()


async def test_master_type_failover_no_masters(minion_opts):
    """
    Tests master_type "failover" to not fall back to 127.0.0.1 address when no master can be resolved
    """
    minion_opts.update(
        {
            "master_type": "failover",
            "master": ["master1", "master2"],
            "__role": "",
            "retry_dns": 0,
        }
    )

    def mock_resolve_dns(opts, fallback=False):
        assert not fallback
        raise SaltClientError("Cannot resolve {}".format(opts["master"]))

    with patch("salt.minion.resolve_dns", mock_resolve_dns), patch(
        "salt.loader.grains", MagicMock(return_value=[])
    ):
        with pytest.raises(SaltClientError):
            minion = salt.minion.Minion(minion_opts)
            # Mock the io_loop so calls to stop/close won't happen.
            minion.io_loop = MagicMock()
            await minion.connect_master()


def test_config_cache_path_overrides():
    cachedir = os.path.abspath("/path/to/master/cache")
    opts = {"cachedir": cachedir, "conf_file": None}

    mminion = salt.minion.MasterMinion(opts)
    assert mminion.opts["cachedir"] == cachedir


def test_minion_grains_refresh_pre_exec_false(minion_opts):
    """
    Minion does not refresh grains when grains_refresh_pre_exec is False
    """
    minion_opts["multiprocessing"] = False
    minion_opts["grains_refresh_pre_exec"] = False
    mock_data = {"fun": "foo.bar", "jid": 123}
    with patch("salt.loader.grains") as grainsfunc, patch(
        "salt.minion.Minion._target", MagicMock(return_value=True)
    ):
        loop = tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=None,
            io_loop=loop,
            load_grains=False,
        )
        try:
            loop.run_sync(lambda: minion._handle_decoded_payload(mock_data))
            grainsfunc.assert_not_called()
        finally:
            minion.destroy()
            loop.close(all_fds=True)


def test_minion_grains_refresh_pre_exec_true(minion_opts):
    """
    Minion refreshes grains when grains_refresh_pre_exec is True
    """
    minion_opts["multiprocessing"] = False
    minion_opts["grains_refresh_pre_exec"] = True
    mock_data = {"fun": "foo.bar", "jid": 123}
    with patch("salt.loader.grains") as grainsfunc, patch(
        "salt.minion.Minion._target", MagicMock(return_value=True)
    ):
        loop = tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=None,
            io_loop=loop,
            load_grains=False,
        )
        try:
            loop.run_sync(lambda: minion._handle_decoded_payload(mock_data))
            grainsfunc.assert_called()
        finally:
            minion.destroy()
            loop.close(all_fds=True)


@pytest.mark.skip_on_darwin(
    reason="Skip on MacOS, where this does not raise an exception."
)
def test_valid_ipv4_master_address_ipv6_enabled(minion_opts):
    """
    Tests that the lookups fail back to ipv4 when ipv6 fails.
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": True,
            "master": "127.0.0.1",
            "master_port": "4555",
            "retry_dns": False,
            "source_address": "111.1.0.1",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        expected = {
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
            "source_ret_port": 49017,
            "master_ip": "127.0.0.1",
        }
        assert salt.minion.resolve_dns(minion_opts) == expected


async def test_master_type_disable(minion_opts):
    """
    Tests master_type "disable" to not even attempt connecting to a master.
    """
    minion_opts.update(
        {
            "master_type": "disable",
            "master": None,
            "__role": "",
            "pub_ret": False,
            "file_client": "local",
        }
    )

    minion = salt.minion.Minion(minion_opts)
    try:

        try:
            minion_man = salt.minion.MinionManager(minion_opts)
            await minion_man._connect_minion(minion)
        except RuntimeError:
            pytest.fail("_connect_minion(minion) threw an error, This was not expected")

        # Make sure beacons and sheduler are initialized
        assert "beacons" in minion.periodic_callbacks
        assert "schedule" in minion.periodic_callbacks
        assert minion.connected is False
    finally:
        # Mock the io_loop so calls to stop/close won't happen.
        minion.io_loop = MagicMock()
        minion.destroy()


async def test_syndic_async_req_channel(syndic_opts):
    syndic_opts["_minion_conf_file"] = ""
    syndic_opts["master_uri"] = "tcp://127.0.0.1:4506"
    syndic = salt.minion.Syndic(syndic_opts)
    syndic.pub_channel = MagicMock()
    syndic.tune_in_no_block()
    assert isinstance(syndic.async_req_channel, salt.channel.client.AsyncReqChannel)


@pytest.mark.slow_test
def test_load_args_and_kwargs(minion_opts):
    """
    Ensure load_args_and_kwargs performs correctly
    """
    _args = [{"max": 40, "__kwarg__": True}]
    ret = salt.minion.load_args_and_kwargs(test_mod.rand_sleep, _args)
    assert ret == ([], {"max": 40})
    assert all([True if "__kwarg__" in item else False for item in _args])

    # Test invalid arguments
    _args = [{"max_sleep": 40, "__kwarg__": True}]
    with pytest.raises(salt.exceptions.SaltInvocationError):
        ret = salt.minion.load_args_and_kwargs(test_mod.rand_sleep, _args)


async def test_connect_master_salt_client_error(minion_opts, connect_master_mock):
    """
    Ensure minion's destory method is called on an salt client error while connecting to master.
    """
    minion_opts["acceptance_wait_time"] = 0
    mm = salt.minion.MinionManager(minion_opts)
    minion = salt.minion.Minion(minion_opts)

    connect_master_mock.exc = SaltClientError
    minion.connect_master = connect_master_mock
    minion.destroy = MagicMock()
    await mm._connect_minion(minion)
    minion.destroy.assert_called_once()

    # The first call raised an error which caused minion.destroy to get called,
    # the second call is a success.
    assert minion.connect_master.calls == 2


async def test_connect_master_unresolveable_error(minion_opts, connect_master_mock):
    """
    Ensure minion's destory method is called on an unresolvable while connecting to master.
    """
    mm = salt.minion.MinionManager(minion_opts)
    minion = salt.minion.Minion(minion_opts)
    connect_master_mock.exc = SaltMasterUnresolvableError
    minion.connect_master = connect_master_mock
    minion.destroy = MagicMock()
    await mm._connect_minion(minion)
    minion.destroy.assert_called_once()

    # Unresolvable errors break out of the loop.
    assert minion.connect_master.calls == 1


async def test_connect_master_general_exception_error(minion_opts, connect_master_mock):
    """
    Ensure minion's destory method is called on an un-handled exception while connecting to master.
    """
    mm = salt.minion.MinionManager(minion_opts)
    minion = salt.minion.Minion(minion_opts)
    connect_master_mock.exc = SaltClientError
    minion.connect_master = connect_master_mock
    minion.destroy = MagicMock()
    await mm._connect_minion(minion)
    minion.destroy.assert_called_once()

    # The first call raised an error which caused minion.destroy to get called,
    # the second call is a success.
    assert minion.connect_master.calls == 2


async def test_minion_manager_async_stop(io_loop, minion_opts, tmp_path):
    """
    Ensure MinionManager's stop method works correctly and calls the
    stop_async method
    """
    # Setup sock_dir with short path
    minion_opts["sock_dir"] = str(tmp_path / "sock")

    os.makedirs(minion_opts["sock_dir"])

    # Create a MinionManager instance with a mock minion
    mm = salt.minion.MinionManager(minion_opts)
    minion = MagicMock(name="minion")
    parent_signal_handler = MagicMock(name="parent_signal_handler")
    mm.minions.append(minion)

    # Set up event publisher and event
    mm._bind()
    assert mm.event_publisher is not None
    assert mm.event is not None

    # Check io_loop is running
    # mm.io_loop is now an asyncio.AbstractEventLoop (not Tornado IOLoop)
    assert mm.io_loop.is_running()

    # Wait for the ipc socket to be created, meaning the publish server is listening.
    while not list(pathlib.Path(minion_opts["sock_dir"]).glob("*")):
        await tornado.gen.sleep(0.3)

    # Set up values for event to send
    load = {"key": "value"}
    ret = {}

    # Connect to minion event bus
    with salt.utils.event.get_event("minion", opts=minion_opts, listen=True) as event:

        # call stop to start stopping the minion
        # mm.stop(signal.SIGTERM, parent_signal_handler)
        mm.stop(signal.SIGTERM, parent_signal_handler)

        # Fire an event and ensure we can still read it back while the minion
        # is stopping
        assert await event.fire_event_async(load, "test_event", timeout=1) is not False
        start = time.monotonic()
        while time.monotonic() - start < 5:
            ret = event.get_event(tag="test_event", wait=1)
            if ret:
                break
            await tornado.gen.sleep(0.3)
    assert "key" in ret
    assert ret["key"] == "value"

    # Sleep to allow stop_async to complete
    await tornado.gen.sleep(5)

    # Ensure stop_async has been called
    minion.destroy.assert_called_once()
    parent_signal_handler.assert_called_once_with(signal.SIGTERM, None)
    assert mm.event_publisher is None
    assert mm.event is None


def test_minion_io_loop_is_asyncio_loop(minion_opts):
    """
    Test that Minion io_loop is converted to asyncio.AbstractEventLoop.
    This verifies the salt.utils.asynchronous.aioloop() conversion.
    """
    minion = salt.minion.Minion(minion_opts, load_grains=False)
    try:
        # Verify io_loop is an asyncio loop, not a Tornado IOLoop
        assert isinstance(minion.io_loop, asyncio.AbstractEventLoop)
        # Ensure it has asyncio methods
        assert hasattr(minion.io_loop, "create_task")
        assert hasattr(minion.io_loop, "call_soon")
        # Ensure it doesn't have Tornado-specific methods
        assert not hasattr(minion.io_loop, "spawn_callback")
    finally:
        minion.destroy()


def test_minion_io_loop_with_provided_loop(minion_opts):
    """
    Test that Minion io_loop conversion works when a loop is provided.
    """
    # Create a Tornado IOLoop
    tornado_loop = tornado.ioloop.IOLoop()
    try:
        minion = salt.minion.Minion(
            minion_opts, io_loop=tornado_loop, load_grains=False
        )
        try:
            # Should still be converted to asyncio loop
            assert isinstance(minion.io_loop, asyncio.AbstractEventLoop)
            # Should be the same underlying loop
            assert minion.io_loop is tornado_loop.asyncio_loop
        finally:
            minion.destroy()
    finally:
        tornado_loop.close()


def test_minion_manager_io_loop_is_asyncio_loop(minion_opts):
    """
    Test that MinionManager io_loop is converted to asyncio.AbstractEventLoop.
    """
    with patch("salt.utils.process.SignalHandlingProcess.start"):
        with patch("salt.utils.verify.valid_id"):
            mm = salt.minion.MinionManager(minion_opts)
            try:
                # Verify io_loop is an asyncio loop
                assert isinstance(mm.io_loop, asyncio.AbstractEventLoop)
                # Ensure it has asyncio methods
                assert hasattr(mm.io_loop, "create_task")
                assert hasattr(mm.io_loop, "call_soon")
                # Ensure it doesn't have Tornado-specific methods
                assert not hasattr(mm.io_loop, "spawn_callback")
            finally:
                mm.destroy()


def test_syndic_manager_io_loop_is_asyncio_loop(minion_opts):
    """
    Test that SyndicManager io_loop is converted to asyncio.AbstractEventLoop.
    """
    minion_opts["order_masters"] = True
    sm = salt.minion.SyndicManager(minion_opts)
    try:
        # Verify io_loop is an asyncio loop
        assert isinstance(sm.io_loop, asyncio.AbstractEventLoop)
        # Ensure it has asyncio methods
        assert hasattr(sm.io_loop, "create_task")
        assert hasattr(sm.io_loop, "call_soon")
        # Ensure it doesn't have Tornado-specific methods
        assert not hasattr(sm.io_loop, "spawn_callback")
    finally:
        sm.destroy()
